"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert = require("assert");
const events_1 = require("events");
const constant_1 = require("./constant");
const finality_manager_1 = require("./finality_manager");
const forkChoiceRule = require("./fork_choice_rule");
const types_1 = require("./types");
const utils_1 = require("./utils");
exports.CONSENSUS_STATE_FINALIZED_HEIGHT_KEY = 'bft:finalizedHeight';
exports.EVENT_BFT_BLOCK_FINALIZED = 'EVENT_BFT_BLOCK_FINALIZED';
class BFT extends events_1.EventEmitter {
    constructor({ chain, dpos, activeDelegates, startingHeight, }) {
        super();
        this._chain = chain;
        this._dpos = dpos;
        this.constants = {
            activeDelegates,
            startingHeight,
        };
    }
    async init(stateStore) {
        this._finalityManager = await this._initFinalityManager(stateStore);
        this.finalityManager.on(finality_manager_1.EVENT_BFT_FINALIZED_HEIGHT_CHANGED, updatedFinalizedHeight => {
            this.emit(finality_manager_1.EVENT_BFT_FINALIZED_HEIGHT_CHANGED, updatedFinalizedHeight);
        });
        const lastBlock = await this._chain.dataAccess.getLastBlockHeader();
        await this.finalityManager.recompute(lastBlock.height, stateStore);
    }
    serialize(blockInstance) {
        return {
            ...blockInstance,
            maxHeightPreviouslyForged: blockInstance.maxHeightPreviouslyForged || 0,
            maxHeightPrevoted: blockInstance.maxHeightPrevoted || 0,
        };
    }
    get finalityManager() {
        return this._finalityManager;
    }
    async deleteBlocks(blocks, stateStore) {
        assert(blocks, 'Must provide blocks which are deleted');
        assert(Array.isArray(blocks), 'Must provide list of blocks');
        const blockHeights = blocks.map(({ height }) => height);
        assert(!blockHeights.some(h => h <= this.finalityManager.finalizedHeight), 'Can not delete block below or same as finalized height');
        const minimumHeight = Math.min(...blockHeights);
        await this.finalityManager.recompute(minimumHeight - 1, stateStore);
    }
    async addNewBlock(block, stateStore) {
        await this.finalityManager.addBlockHeader(block, stateStore);
        const { finalizedHeight } = this.finalityManager;
        stateStore.consensus.set(exports.CONSENSUS_STATE_FINALIZED_HEIGHT_KEY, String(finalizedHeight));
    }
    async verifyNewBlock(blockHeader) {
        const bftHeaders = await this.finalityManager.getBFTApplicableBlockHeaders(blockHeader.height - 1);
        return this.finalityManager.verifyBlockHeaders(blockHeader, bftHeaders);
    }
    forkChoice(block, lastBlock) {
        const receivedBlock = {
            ...block,
            receivedAt: this._chain.slots.getEpochTime(),
        };
        if (forkChoiceRule.isValidBlock(lastBlock, receivedBlock)) {
            return types_1.ForkStatus.VALID_BLOCK;
        }
        if (forkChoiceRule.isIdenticalBlock(lastBlock, receivedBlock)) {
            return types_1.ForkStatus.IDENTICAL_BLOCK;
        }
        if (forkChoiceRule.isDoubleForging(lastBlock, receivedBlock)) {
            return types_1.ForkStatus.DOUBLE_FORGING;
        }
        if (forkChoiceRule.isTieBreak({
            slots: this._chain.slots,
            lastAppliedBlock: lastBlock,
            receivedBlock,
        })) {
            return types_1.ForkStatus.TIE_BREAK;
        }
        if (forkChoiceRule.isDifferentChain(lastBlock, receivedBlock)) {
            return types_1.ForkStatus.DIFFERENT_CHAIN;
        }
        return types_1.ForkStatus.DISCARD;
    }
    async _initFinalityManager(stateStore) {
        const storedFinalizedHeight = await stateStore.consensus.get(exports.CONSENSUS_STATE_FINALIZED_HEIGHT_KEY);
        const finalizedHeightStored = storedFinalizedHeight
            ? parseInt(storedFinalizedHeight, 10)
            : 1;
        const bftMigrationHeight = this.constants.startingHeight -
            this.constants.activeDelegates * constant_1.BFT_MIGRATION_ROUND_OFFSET;
        const finalizedHeight = Math.max(finalizedHeightStored, bftMigrationHeight);
        return new finality_manager_1.FinalityManager({
            chain: this._chain,
            dpos: this._dpos,
            finalizedHeight,
            activeDelegates: this.constants.activeDelegates,
        });
    }
    async isBFTProtocolCompliant(blockHeader) {
        assert(blockHeader, 'No block was provided to be verified');
        const roundsThreshold = 3;
        const heightThreshold = this.constants.activeDelegates * roundsThreshold;
        if (blockHeader.maxHeightPreviouslyForged === 0) {
            return true;
        }
        const bftHeaders = await this.finalityManager.getBFTApplicableBlockHeaders(blockHeader.height - 1);
        const maxHeightPreviouslyForgedBlock = bftHeaders.find(bftHeader => bftHeader.height === blockHeader.maxHeightPreviouslyForged);
        if (!maxHeightPreviouslyForgedBlock ||
            blockHeader.maxHeightPreviouslyForged >= blockHeader.height ||
            (blockHeader.height - blockHeader.maxHeightPreviouslyForged <=
                heightThreshold &&
                blockHeader.generatorPublicKey !==
                    maxHeightPreviouslyForgedBlock.generatorPublicKey)) {
            return false;
        }
        return true;
    }
    get finalizedHeight() {
        return this.finalityManager.finalizedHeight;
    }
    get maxHeightPrevoted() {
        return this.finalityManager.chainMaxHeightPrevoted;
    }
    reset() {
        this.finalityManager.reset();
    }
    validateBlock(block) {
        utils_1.validateBlockHeader(block);
    }
}
exports.BFT = BFT;
//# sourceMappingURL=bft.js.map