"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const assert = require("assert");
const Debug = require("debug");
const events_1 = require("events");
const constant_1 = require("./constant");
const types_1 = require("./types");
const utils_1 = require("./utils");
const debug = Debug('lisk:bft:consensus_manager');
exports.EVENT_BFT_FINALIZED_HEIGHT_CHANGED = 'EVENT_BFT_FINALIZED_HEIGHT_CHANGED';
class FinalityManager extends events_1.EventEmitter {
    constructor({ chain, dpos, finalizedHeight, activeDelegates, }) {
        super();
        assert(activeDelegates > 0, 'Must provide a positive activeDelegates');
        this._chain = chain;
        this._dpos = dpos;
        this.activeDelegates = activeDelegates;
        this.preVoteThreshold = Math.ceil((this.activeDelegates * 2) / 3);
        this.preCommitThreshold = Math.ceil((this.activeDelegates * 2) / 3);
        this.processingThreshold = this.activeDelegates * constant_1.BFT_ROUND_THRESHOLD - 1;
        this.maxHeaders = this.activeDelegates * 5;
        this.finalizedHeight = finalizedHeight;
        this.chainMaxHeightPrevoted = 0;
        this.state = {};
        this.preVotes = {};
        this.preCommits = {};
    }
    async addBlockHeader(blockHeader, stateStore) {
        debug('addBlockHeader invoked');
        debug('validateBlockHeader invoked');
        const bftApplicableBlocks = await this.getBFTApplicableBlockHeaders(blockHeader.height - 1);
        utils_1.validateBlockHeader(blockHeader);
        this.verifyBlockHeaders(blockHeader, bftApplicableBlocks);
        await this.updatePreVotesPreCommits(blockHeader, stateStore, bftApplicableBlocks);
        this.updatePreVotedAndFinalizedHeight();
        this._cleanup();
        debug('after adding block header', {
            finalizedHeight: this.finalizedHeight,
            chainMaxHeightPrevoted: this.chainMaxHeightPrevoted,
        });
        return this;
    }
    async updatePreVotesPreCommits(header, stateStore, bftBlockHeaders) {
        debug('updatePreVotesPreCommits invoked');
        if (header.maxHeightPreviouslyForged >= header.height) {
            return false;
        }
        const { generatorPublicKey: delegatePublicKey } = header;
        const delegateAddress = lisk_cryptography_1.getAddressFromPublicKey(delegatePublicKey);
        const isStandby = await this._dpos.isStandbyDelegate(delegateAddress, header.height, stateStore);
        if (isStandby) {
            return false;
        }
        const delegateState = this.state[delegatePublicKey] || {
            maxPreVoteHeight: 0,
            maxPreCommitHeight: 0,
        };
        const minValidHeightToPreCommit = this._getMinValidHeightToPreCommit(header, bftBlockHeaders);
        const delegateMinHeightActive = await this._dpos.getMinActiveHeight(header.height, delegateAddress, stateStore);
        const minPreCommitHeight = Math.max(header.height - this.processingThreshold, delegateMinHeightActive, minValidHeightToPreCommit, delegateState.maxPreCommitHeight + 1);
        const maxPreCommitHeight = header.height - 1;
        for (let j = minPreCommitHeight; j <= maxPreCommitHeight; j += 1) {
            if (this.preVotes[j] >= this.preVoteThreshold) {
                this.preCommits[j] = (this.preCommits[j] || 0) + 1;
                delegateState.maxPreCommitHeight = j;
            }
        }
        const minPreVoteHeight = Math.max(delegateMinHeightActive, header.maxHeightPreviouslyForged + 1, header.height - this.processingThreshold);
        const maxPreVoteHeight = header.height;
        for (let j = minPreVoteHeight; j <= maxPreVoteHeight; j += 1) {
            this.preVotes[j] = (this.preVotes[j] || 0) + 1;
        }
        delegateState.maxPreVoteHeight = maxPreVoteHeight;
        this.state[delegatePublicKey] = delegateState;
        return true;
    }
    updatePreVotedAndFinalizedHeight() {
        debug('updatePreVotedAndFinalizedHeight invoked');
        const highestHeightPreVoted = Object.keys(this.preVotes)
            .reverse()
            .find(key => this.preVotes[key] >= this.preVoteThreshold);
        this.chainMaxHeightPrevoted = highestHeightPreVoted
            ? parseInt(highestHeightPreVoted, 10)
            : this.chainMaxHeightPrevoted;
        const highestHeightPreCommitted = Object.keys(this.preCommits)
            .reverse()
            .find(key => this.preCommits[key] >= this.preCommitThreshold);
        const previouslyFinalizedHeight = this.finalizedHeight;
        if (highestHeightPreCommitted) {
            this.finalizedHeight = parseInt(highestHeightPreCommitted, 10);
        }
        if (previouslyFinalizedHeight !== this.finalizedHeight) {
            this.emit(exports.EVENT_BFT_FINALIZED_HEIGHT_CHANGED, this.finalizedHeight);
        }
        return true;
    }
    _getMinValidHeightToPreCommit(header, bftApplicableBlocks) {
        let needleHeight = Math.max(header.maxHeightPreviouslyForged, header.height - this.processingThreshold);
        const searchTillHeight = Math.max(1, header.height - this.processingThreshold);
        let previousBlockHeight = header.maxHeightPreviouslyForged;
        const blocksIncludingCurrent = [header, ...bftApplicableBlocks];
        while (needleHeight >= searchTillHeight) {
            if (needleHeight === previousBlockHeight) {
                const previousBlockHeader = blocksIncludingCurrent.find(bftHeader => bftHeader.height === needleHeight);
                if (!previousBlockHeader) {
                    debug('Fail to get cached block header');
                    return 0;
                }
                if (previousBlockHeader.generatorPublicKey !==
                    header.generatorPublicKey ||
                    previousBlockHeader.maxHeightPreviouslyForged >= needleHeight) {
                    return needleHeight + 1;
                }
                previousBlockHeight = previousBlockHeader.maxHeightPreviouslyForged;
                needleHeight = previousBlockHeader.maxHeightPreviouslyForged;
            }
            else {
                needleHeight -= 1;
            }
        }
        return Math.max(needleHeight + 1, searchTillHeight);
    }
    async recompute(recomputeUptoHeight, stateStore) {
        this.reset();
        const blockHeaders = await this.getBFTApplicableBlockHeaders(recomputeUptoHeight);
        if (!blockHeaders.length) {
            throw new Error('Cannot find a block to recompute');
        }
        for (let i = blockHeaders.length - 1; i >= 0; i -= 1) {
            const blockHeader = blockHeaders[i];
            const bftBlockHeader = blockHeaders.slice(i + 1);
            await this.updatePreVotesPreCommits(blockHeader, stateStore, bftBlockHeader);
        }
        this.updatePreVotedAndFinalizedHeight();
        this._cleanup();
    }
    reset() {
        this.state = {};
        this.chainMaxHeightPrevoted = 0;
        this.preVotes = {};
        this.preCommits = {};
    }
    verifyBlockHeaders(blockHeader, bftBlockHeaders) {
        debug('verifyBlockHeaders invoked');
        debug(blockHeader);
        if (bftBlockHeaders.length >= this.processingThreshold &&
            blockHeader.maxHeightPrevoted !== this.chainMaxHeightPrevoted) {
            throw new types_1.BFTInvalidAttributeError(`Wrong maxHeightPrevoted in blockHeader. maxHeightPrevoted: ${blockHeader.maxHeightPrevoted}, : ${this.chainMaxHeightPrevoted}`);
        }
        const delegateLastBlock = bftBlockHeaders.find(header => header.generatorPublicKey === blockHeader.generatorPublicKey);
        if (!delegateLastBlock) {
            return true;
        }
        let earlierBlock = delegateLastBlock;
        let laterBlock = blockHeader;
        const higherMaxHeightPreviouslyForged = earlierBlock.maxHeightPreviouslyForged >
            laterBlock.maxHeightPreviouslyForged;
        const sameMaxHeightPreviouslyForged = earlierBlock.maxHeightPreviouslyForged ===
            laterBlock.maxHeightPreviouslyForged;
        const higherMaxHeightPrevoted = earlierBlock.maxHeightPrevoted > laterBlock.maxHeightPrevoted;
        const sameMaxHeightPrevoted = earlierBlock.maxHeightPrevoted === laterBlock.maxHeightPrevoted;
        const higherHeight = earlierBlock.height > laterBlock.height;
        if (higherMaxHeightPreviouslyForged ||
            (sameMaxHeightPreviouslyForged && higherMaxHeightPrevoted) ||
            (sameMaxHeightPreviouslyForged && sameMaxHeightPrevoted && higherHeight)) {
            [earlierBlock, laterBlock] = [laterBlock, earlierBlock];
        }
        if (earlierBlock.maxHeightPrevoted === laterBlock.maxHeightPrevoted &&
            earlierBlock.height >= laterBlock.height) {
            throw new types_1.BFTForkChoiceRuleError();
        }
        if (earlierBlock.height > laterBlock.maxHeightPreviouslyForged) {
            throw new types_1.BFTChainDisjointError();
        }
        if (earlierBlock.maxHeightPrevoted > laterBlock.maxHeightPrevoted) {
            throw new types_1.BFTLowerChainBranchError();
        }
        return true;
    }
    async getBFTApplicableBlockHeaders(height) {
        if (height === 0) {
            return [];
        }
        const fromHeight = Math.max(1, height - this.processingThreshold);
        return this._chain.dataAccess.getBlockHeadersByHeightBetween(fromHeight, height);
    }
    _cleanup() {
        Object.keys(this.preVotes)
            .slice(0, this.maxHeaders * -1)
            .forEach(key => {
            delete this.preVotes[key];
        });
        Object.keys(this.preCommits)
            .slice(0, this.maxHeaders * -1)
            .forEach(key => {
            delete this.preCommits[key];
        });
    }
}
exports.FinalityManager = FinalityManager;
//# sourceMappingURL=finality_manager.js.map