"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const Debug = require("debug");
const events_1 = require("events");
const block_reward_1 = require("./block_reward");
const constants_1 = require("./constants");
const data_access_1 = require("./data_access");
const slots_1 = require("./slots");
const state_store_1 = require("./state_store");
const transactions_1 = require("./transactions");
const blocksUtils = require("./utils");
const validate_1 = require("./validate");
const verify_1 = require("./verify");
const TRANSACTION_TYPES_VOTE = [3, 11];
const saveBlock = async (storage, blockJSON, tx) => {
    if (!tx) {
        throw new Error('Block should only be saved in a database tx');
    }
    const promises = [storage.entities.Block.create(blockJSON, {}, tx)];
    if (blockJSON.transactions.length) {
        promises.push(storage.entities.Transaction.create(blockJSON.transactions, {}, tx));
    }
    return tx.batch(promises);
};
const applyConfirmedStep = async (blockInstance, stateStore) => {
    if (blockInstance.transactions.length <= 0) {
        return;
    }
    const transactionsResponses = await transactions_1.applyTransactions()(blockInstance.transactions, stateStore);
    const unappliableTransactionsResponse = transactionsResponses.filter(transactionResponse => transactionResponse.status !== lisk_transactions_1.Status.OK);
    if (unappliableTransactionsResponse.length > 0) {
        throw unappliableTransactionsResponse[0].errors;
    }
};
const applyConfirmedGenesisStep = async (blockInstance, stateStore) => {
    blockInstance.transactions.sort(a => {
        if (TRANSACTION_TYPES_VOTE.includes(a.type)) {
            return 1;
        }
        return 0;
    });
    const sortedTransactionInstances = [...blockInstance.transactions];
    await transactions_1.applyGenesisTransactions()(sortedTransactionInstances, stateStore);
    return blockInstance;
};
const undoConfirmedStep = async (blockInstance, stateStore) => {
    if (blockInstance.transactions.length === 0) {
        return;
    }
    const transactionsResponses = await transactions_1.undoTransactions()(blockInstance.transactions, stateStore);
    const unappliedTransactionResponse = transactionsResponses.find(transactionResponse => transactionResponse.status !== lisk_transactions_1.Status.OK);
    if (unappliedTransactionResponse) {
        throw unappliedTransactionResponse.errors;
    }
};
const debug = Debug('lisk:chain');
class Chain {
    constructor({ storage, genesisBlock, registeredTransactions, epochTime, blockTime, networkIdentifier, maxPayloadLength, rewardDistance, rewardOffset, rewardMilestones, totalAmount, stateBlockSize = constants_1.DEFAULT_STATE_BLOCK_SIZE, minBlockHeaderCache = constants_1.DEFAULT_MIN_BLOCK_HEADER_CACHE, maxBlockHeaderCache = constants_1.DEFAULT_MAX_BLOCK_HEADER_CACHE, }) {
        this.events = new events_1.EventEmitter();
        this.storage = storage;
        this.dataAccess = new data_access_1.DataAccess({
            dbStorage: storage,
            registeredTransactions,
            minBlockHeaderCache,
            maxBlockHeaderCache,
        });
        const genesisInstance = this.dataAccess.deserialize(genesisBlock);
        this._lastBlock = genesisInstance;
        this._networkIdentifier = networkIdentifier;
        this.genesisBlock = genesisInstance;
        this.slots = new slots_1.Slots({ epochTime, interval: blockTime });
        this.blockRewardArgs = {
            distance: rewardDistance,
            rewardOffset,
            milestones: rewardMilestones,
            totalAmount,
        };
        this.blockReward = {
            calculateMilestone: height => block_reward_1.calculateMilestone(height, this.blockRewardArgs),
            calculateReward: height => block_reward_1.calculateReward(height, this.blockRewardArgs),
            calculateSupply: height => block_reward_1.calculateSupply(height, this.blockRewardArgs),
        };
        this.constants = {
            stateBlockSize,
            epochTime,
            blockTime,
            maxPayloadLength,
        };
        this.blocksVerify = new verify_1.BlocksVerify({
            dataAccess: this.dataAccess,
            genesisBlock: this.genesisBlock,
        });
    }
    get lastBlock() {
        const { receivedAt, ...block } = this._lastBlock;
        return block;
    }
    deserialize(blockJSON) {
        return this.dataAccess.deserialize(blockJSON);
    }
    serialize(blockJSON) {
        return this.dataAccess.serialize(blockJSON);
    }
    serializeBlockHeader(blockHeader) {
        return this.dataAccess.serializeBlockHeader(blockHeader);
    }
    deserializeBlockHeader(blockJSON) {
        return this.dataAccess.deserializeBlockHeader(blockJSON);
    }
    deserializeTransaction(transactionJSON) {
        return this.dataAccess.deserializeTransaction(transactionJSON);
    }
    async init() {
        const genesisBlock = await this.dataAccess.getBlockHeaderByHeight(1);
        if (!genesisBlock) {
            throw new Error('Failed to load genesis block');
        }
        const genesisBlockMatch = this.blocksVerify.matchGenesisBlock(genesisBlock);
        if (!genesisBlockMatch) {
            throw new Error('Genesis block does not match');
        }
        const storageLastBlock = await this.dataAccess.getLastBlock();
        if (!storageLastBlock) {
            throw new Error('Failed to load last block');
        }
        if (storageLastBlock.height !== genesisBlock.height) {
            await this._cacheBlockHeaders(storageLastBlock);
        }
        this._lastBlock = storageLastBlock;
    }
    resetBlockHeaderCache() {
        this.dataAccess.resetBlockHeaderCache();
    }
    async newStateStore(skipLastHeights = 0) {
        var _a, _b;
        const fromHeight = Math.max(1, this._lastBlock.height - this.constants.stateBlockSize - skipLastHeights);
        const toHeight = Math.max(this._lastBlock.height - skipLastHeights, 1);
        const lastBlockHeaders = await this.dataAccess.getBlockHeadersByHeightBetween(fromHeight, toHeight);
        const lastBlockReward = this.blockReward.calculateReward((_b = (_a = lastBlockHeaders[0]) === null || _a === void 0 ? void 0 : _a.height) !== null && _b !== void 0 ? _b : 1);
        return new state_store_1.StateStore(this.storage, {
            networkIdentifier: this._networkIdentifier,
            lastBlockHeaders,
            lastBlockReward,
        });
    }
    async _cacheBlockHeaders(storageLastBlock) {
        const fromHeight = Math.max(storageLastBlock.height - constants_1.DEFAULT_MAX_BLOCK_HEADER_CACHE, 1);
        const toHeight = storageLastBlock.height;
        debug({ h: storageLastBlock.height, fromHeight, toHeight }, 'Cache block headers during chain init');
        const blockHeaders = await this.dataAccess.getBlockHeadersByHeightBetween(fromHeight, toHeight);
        const sortedBlockHeaders = [...blockHeaders].sort((a, b) => a.height - b.height);
        for (const blockHeader of sortedBlockHeaders) {
            debug({ height: blockHeader.height }, 'Add block header to cache');
            this.dataAccess.addBlockHeader(blockHeader);
        }
    }
    validateBlockHeader(block, blockBytes) {
        validate_1.validatePreviousBlockProperty(block, this.genesisBlock);
        validate_1.validateSignature(block, blockBytes, this._networkIdentifier);
        validate_1.validateReward(block, this.blockReward.calculateReward(block.height));
        const transactionsResponses = transactions_1.validateTransactions()(block.transactions);
        const invalidTransactionResponse = transactionsResponses.find(transactionResponse => transactionResponse.status !== lisk_transactions_1.Status.OK);
        if (invalidTransactionResponse) {
            throw invalidTransactionResponse.errors;
        }
        validate_1.validatePayload(block, this.constants.maxPayloadLength);
        block.id = blocksUtils.getId(blockBytes);
    }
    async resetState() {
        await this.storage.entities.Account.resetMemTables();
        await this.storage.entities.ConsensusState.delete();
        this.dataAccess.resetBlockHeaderCache();
    }
    async verify(blockInstance, _, { skipExistingCheck }) {
        verify_1.verifyPreviousBlockId(blockInstance, this._lastBlock, this.genesisBlock);
        validate_1.validateBlockSlot(blockInstance, this._lastBlock, this.slots);
        if (!skipExistingCheck) {
            await verify_1.verifyBlockNotExists(this.storage, blockInstance);
            const transactionsResponses = await transactions_1.checkPersistedTransactions(this.dataAccess)(blockInstance.transactions);
            const invalidPersistedResponse = transactionsResponses.find(transactionResponse => transactionResponse.status !== lisk_transactions_1.Status.OK);
            if (invalidPersistedResponse) {
                throw invalidPersistedResponse.errors;
            }
        }
        await this.blocksVerify.checkTransactions(blockInstance);
    }
    async apply(blockInstance, stateStore) {
        await applyConfirmedStep(blockInstance, stateStore);
        await block_reward_1.applyFeeAndRewards(blockInstance, stateStore);
    }
    async applyGenesis(blockInstance, stateStore) {
        await applyConfirmedGenesisStep(blockInstance, stateStore);
        await block_reward_1.applyFeeAndRewards(blockInstance, stateStore);
    }
    async save(blockInstance, stateStore, { saveOnlyState, removeFromTempTable } = {
        saveOnlyState: false,
        removeFromTempTable: false,
    }) {
        return this.storage.entities.Block.begin('saveBlock', async (tx) => {
            await stateStore.finalize(tx);
            if (!saveOnlyState) {
                const blockJSON = this.serialize(blockInstance);
                await saveBlock(this.storage, blockJSON, tx);
            }
            if (removeFromTempTable) {
                await this.removeBlockFromTempTable(blockInstance.id, tx);
            }
            this.dataAccess.addBlockHeader(blockInstance);
            this._lastBlock = blockInstance;
            const accounts = stateStore.account
                .getUpdated()
                .map(anAccount => anAccount.toJSON());
            this.events.emit(constants_1.EVENT_NEW_BLOCK, {
                block: this.serialize(blockInstance),
                accounts,
            });
        });
    }
    async undo(blockInstance, stateStore) {
        await block_reward_1.undoFeeAndRewards(blockInstance, stateStore);
        await undoConfirmedStep(blockInstance, stateStore);
    }
    async _deleteLastBlock(lastBlock, tx) {
        if (lastBlock.height === 1) {
            throw new Error('Cannot delete genesis block');
        }
        const block = await this.dataAccess.getBlockByID(lastBlock.previousBlockId);
        if (!block) {
            throw new Error('PreviousBlock is null');
        }
        await this.storage.entities.Block.delete({ id: lastBlock.id }, {}, tx);
        return block;
    }
    async remove(block, stateStore, { saveTempBlock } = { saveTempBlock: false }) {
        await this.storage.entities.Block.begin('revertBlock', async (tx) => {
            const secondLastBlock = await this._deleteLastBlock(block, tx);
            if (saveTempBlock) {
                const blockJSON = this.serialize(block);
                const blockTempEntry = {
                    id: blockJSON.id,
                    height: blockJSON.height,
                    fullBlock: blockJSON,
                };
                await this.storage.entities.TempBlock.create(blockTempEntry, {}, tx);
            }
            await stateStore.finalize(tx);
            await this.dataAccess.removeBlockHeader(block.id);
            this._lastBlock = secondLastBlock;
            const accounts = stateStore.account
                .getUpdated()
                .map(anAccount => anAccount.toJSON());
            this.events.emit(constants_1.EVENT_DELETE_BLOCK, {
                block: this.serialize(block),
                accounts,
            });
        });
    }
    async removeBlockFromTempTable(blockId, tx) {
        return this.storage.entities.TempBlock.delete({ id: blockId }, {}, tx);
    }
    async exists(block) {
        try {
            await verify_1.verifyBlockNotExists(this.storage, block);
            return false;
        }
        catch (err) {
            return true;
        }
    }
    async getHighestCommonBlock(ids) {
        const blocks = await this.dataAccess.getBlockHeadersByIDs(ids);
        const sortedBlocks = [...blocks].sort((a, b) => b.height - a.height);
        const highestCommonBlock = sortedBlocks.shift();
        return highestCommonBlock;
    }
    async filterReadyTransactions(transactions, context) {
        const stateStore = await this.newStateStore();
        const allowedTransactionsIds = transactions_1.checkAllowedTransactions(context)(transactions)
            .filter(transactionResponse => transactionResponse.status === lisk_transactions_1.Status.OK)
            .map(transactionResponse => transactionResponse.id);
        const allowedTransactions = transactions.filter(transaction => allowedTransactionsIds.includes(transaction.id));
        const transactionsResponses = await transactions_1.applyTransactions()(allowedTransactions, stateStore);
        const readyTransactions = allowedTransactions.filter(transaction => transactionsResponses
            .filter(response => response.status === lisk_transactions_1.Status.OK)
            .map(response => response.id)
            .includes(transaction.id));
        return readyTransactions;
    }
    async validateTransactions(transactions) {
        return transactions_1.composeTransactionSteps(transactions_1.checkAllowedTransactions({
            blockVersion: this.lastBlock.version,
            blockHeight: this.lastBlock.height,
            blockTimestamp: this.lastBlock.timestamp,
        }), transactions_1.validateTransactions())(transactions);
    }
    async applyTransactions(transactions) {
        const stateStore = await this.newStateStore();
        return transactions_1.composeTransactionSteps(transactions_1.checkAllowedTransactions(() => {
            const { version, height, timestamp } = this._lastBlock;
            return {
                blockVersion: version,
                blockHeight: height,
                blockTimestamp: timestamp,
            };
        }), transactions_1.checkPersistedTransactions(this.dataAccess), transactions_1.applyTransactions())(transactions, stateStore);
    }
    async applyTransactionsWithStateStore(transactions, stateStore) {
        return transactions_1.composeTransactionSteps(transactions_1.checkPersistedTransactions(this.dataAccess), transactions_1.applyTransactions())(transactions, stateStore);
    }
    getTotalEarningAndBurnt(blockInstance) {
        const { totalFee, totalMinFee } = block_reward_1.getTotalFees(blockInstance);
        return {
            totalEarning: blockInstance.reward + totalFee - totalMinFee,
            totalBurnt: totalMinFee,
        };
    }
}
exports.Chain = Chain;
//# sourceMappingURL=chain.js.map