"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
exports.validateSignature = (block, blockBytes, networkIdentifier) => {
    const signatureLength = 64;
    const dataWithoutSignature = blockBytes.slice(0, blockBytes.length - signatureLength);
    const hashedBlock = lisk_cryptography_1.hash(Buffer.concat([
        Buffer.from(networkIdentifier, 'hex'),
        dataWithoutSignature,
    ]));
    const verified = lisk_cryptography_1.verifyData(hashedBlock, block.blockSignature, block.generatorPublicKey);
    if (!verified) {
        throw new Error('Invalid block signature');
    }
};
exports.validatePreviousBlockProperty = (block, genesisBlock) => {
    const isGenesisBlock = block.id === genesisBlock.id &&
        !block.previousBlockId &&
        block.height === 1;
    const propertyIsValid = isGenesisBlock ||
        (block.id !== genesisBlock.id &&
            block.previousBlockId &&
            block.height !== 1);
    if (!propertyIsValid) {
        throw new Error('Invalid previous block');
    }
};
exports.validateReward = (block, maxReward) => {
    if (block.reward > maxReward) {
        throw new Error(`Invalid block reward: ${block.reward.toString()} maximum allowed: ${maxReward.toString()}`);
    }
};
exports.validatePayload = (block, maxPayloadLength) => {
    if (block.payloadLength > maxPayloadLength) {
        throw new Error('Payload length is too long');
    }
    let totalAmount = BigInt(0);
    let totalFee = BigInt(0);
    const transactionsBytesArray = [];
    const appliedTransactions = {};
    block.transactions.forEach(transaction => {
        const transactionBytes = transaction.getBytes();
        if (appliedTransactions[transaction.id]) {
            throw new Error(`Encountered duplicate transaction: ${transaction.id}`);
        }
        appliedTransactions[transaction.id] = transaction;
        if (transactionBytes) {
            transactionsBytesArray.push(transactionBytes);
        }
        totalAmount = totalAmount + BigInt(transaction.asset.amount || 0);
        totalFee = totalFee + BigInt(transaction.fee);
    });
    const transactionsBuffer = Buffer.concat(transactionsBytesArray);
    const payloadHash = lisk_cryptography_1.hash(transactionsBuffer).toString('hex');
    if (payloadHash !== block.payloadHash) {
        throw new Error('Invalid payload hash');
    }
    if (totalAmount !== BigInt(block.totalAmount)) {
        throw new Error('Invalid total amount');
    }
    if (totalFee !== block.totalFee) {
        throw new Error('Invalid total fee');
    }
};
exports.validateBlockSlot = (block, lastBlock, slots) => {
    const blockSlotNumber = slots.getSlotNumber(block.timestamp);
    const lastBlockSlotNumber = slots.getSlotNumber(lastBlock.timestamp);
    if (blockSlotNumber > slots.getSlotNumber() ||
        blockSlotNumber <= lastBlockSlotNumber) {
        throw new Error('Invalid block timestamp');
    }
};
//# sourceMappingURL=validate.js.map