"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const transactionsModule = require("./transactions");
exports.verifyBlockNotExists = async (storage, block) => {
    const isPersisted = await storage.entities.Block.isPersisted({
        id: block.id,
    });
    if (isPersisted) {
        throw new Error(`Block ${block.id} already exists`);
    }
};
exports.verifyPreviousBlockId = (block, lastBlock, genesisBlock) => {
    const isGenesisBlock = block.id === genesisBlock.id &&
        !block.previousBlockId &&
        block.height === 1;
    const isConsecutiveBlock = lastBlock.height + 1 === block.height &&
        block.previousBlockId === lastBlock.id;
    if (!isGenesisBlock && !isConsecutiveBlock) {
        throw new Error('Invalid previous block');
    }
};
class BlocksVerify {
    constructor({ dataAccess, genesisBlock }) {
        this.dataAccess = dataAccess;
        this.genesisBlock = genesisBlock;
    }
    async checkExists(block) {
        const isPersisted = await this.dataAccess.isBlockPersisted(block.id);
        if (isPersisted) {
            throw new Error(`Block ${block.id} already exists`);
        }
        if (!block.transactions.length) {
            return;
        }
        const transactionIDs = block.transactions.map(transaction => transaction.id);
        const persistedTransactions = await this.dataAccess.getTransactionsByIDs(transactionIDs);
        if (persistedTransactions.length > 0) {
            throw new Error(`Transaction is already confirmed: ${persistedTransactions[0].id}`);
        }
    }
    async checkTransactions(blockInstance) {
        const { version, height, timestamp, transactions } = blockInstance;
        if (transactions.length === 0) {
            return;
        }
        const context = {
            blockVersion: version,
            blockHeight: height,
            blockTimestamp: timestamp,
        };
        const nonAllowedTxResponses = transactionsModule
            .checkAllowedTransactions(context)(transactions)
            .find((transactionResponse) => transactionResponse.status !== lisk_transactions_1.Status.OK);
        if (nonAllowedTxResponses) {
            throw nonAllowedTxResponses.errors;
        }
    }
    matchGenesisBlock(block) {
        return (block.id === this.genesisBlock.id &&
            block.payloadHash === this.genesisBlock.payloadHash &&
            block.blockSignature === this.genesisBlock.blockSignature);
    }
}
exports.BlocksVerify = BlocksVerify;
//# sourceMappingURL=verify.js.map