"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const Debug = require("debug");
const constants_1 = require("./constants");
const delegates_list_1 = require("./delegates_list");
const random_seed_1 = require("./random_seed");
const debug = Debug('lisk:dpos:delegate_info');
const _isGenesisBlock = (block) => block.height === 1;
const zeroRandomSeed = Buffer.from('00000000000000000000000000000000', 'hex');
class DelegatesInfo {
    constructor({ rounds, chain, activeDelegates, standbyDelegates, events, delegatesList, }) {
        this.chain = chain;
        this.rounds = rounds;
        this.activeDelegates = activeDelegates;
        this.standbyDelegates = standbyDelegates;
        this.events = events;
        this.delegatesList = delegatesList;
    }
    async apply(block, stateStore, { delegateListRoundOffset }) {
        const undo = false;
        return this._update(block, stateStore, { undo, delegateListRoundOffset });
    }
    async undo(block, stateStore, { delegateListRoundOffset }) {
        const undo = true;
        if (_isGenesisBlock(block)) {
            throw new Error('Cannot undo genesis block');
        }
        return this._update(block, stateStore, { undo, delegateListRoundOffset });
    }
    async _update(block, stateStore, { delegateListRoundOffset, undo }) {
        if (_isGenesisBlock(block)) {
            const initialRound = 1;
            for (let i = initialRound; i <= initialRound + delegateListRoundOffset; i += 1) {
                await this.delegatesList.createVoteWeightsSnapshot(1, stateStore, i - 1);
            }
            await this.delegatesList.updateForgersList(initialRound, [zeroRandomSeed, zeroRandomSeed], stateStore);
            return false;
        }
        const round = this.rounds.calcRound(block.height);
        if (!this._isLastBlockOfTheRound(block)) {
            return false;
        }
        await this._updateMissedBlocks(block, stateStore, undo);
        if (undo) {
            const previousRound = round + 1;
            this.events.emit(constants_1.EVENT_ROUND_CHANGED, {
                oldRound: previousRound,
                newRound: round,
            });
            debug('Deleting delegate list after ', round + delegateListRoundOffset);
            await delegates_list_1.deleteForgersListAfterRound(round, stateStore);
            await delegates_list_1.deleteVoteWeightsAfterRound(round + delegateListRoundOffset, stateStore);
        }
        else {
            const nextRound = round + 1;
            this.events.emit(constants_1.EVENT_ROUND_CHANGED, {
                oldRound: round,
                newRound: nextRound,
            });
            debug('Creating delegate list for', round + delegateListRoundOffset);
            await this.delegatesList.createVoteWeightsSnapshot(block.height + 1, stateStore);
            const [randomSeed1, randomSeed2] = random_seed_1.generateRandomSeeds(round, this.rounds, stateStore.consensus.lastBlockHeaders);
            await this.delegatesList.updateForgersList(nextRound, [randomSeed1, randomSeed2], stateStore);
        }
        return true;
    }
    async _updateMissedBlocks(blockHeader, stateStore, undo) {
        const round = this.rounds.calcRound(blockHeader.height);
        debug('Calculating missed block', round);
        const expectedForgingAddresses = await delegates_list_1.getForgerAddressesForRound(round, stateStore);
        const heightFrom = this.rounds.calcRoundStartHeight(round);
        const heightTo = this.rounds.calcRoundEndHeight(round) - 1;
        const blocksInRounds = await this.chain.dataAccess.getBlockHeadersByHeightBetween(heightFrom, heightTo);
        blocksInRounds.push(blockHeader);
        if (blocksInRounds.length !==
            this.activeDelegates + this.standbyDelegates) {
            throw new Error('Fetched blocks do not match the size of the active delegates');
        }
        const forgedPublicKeys = blocksInRounds.map(block => block.generatorPublicKey);
        const missedBlocksDelegateAddresses = expectedForgingAddresses.filter(expectedAddress => !forgedPublicKeys.find(publicKey => lisk_cryptography_1.getAddressFromPublicKey(publicKey) === expectedAddress));
        if (!missedBlocksDelegateAddresses.length) {
            return;
        }
        for (const address of missedBlocksDelegateAddresses) {
            const account = await stateStore.account.get(address);
            account.missedBlocks += undo ? -1 : 1;
            stateStore.account.set(address, account);
        }
    }
    _isLastBlockOfTheRound(block) {
        const round = this.rounds.calcRound(block.height);
        const nextRound = this.rounds.calcRound(block.height + 1);
        return round < nextRound;
    }
}
exports.DelegatesInfo = DelegatesInfo;
//# sourceMappingURL=delegates_info.js.map