"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const Debug = require("debug");
const constants_1 = require("./constants");
const debug = Debug('lisk:dpos:delegate_list');
const SIZE_UINT64 = 8;
exports.getForgersList = async (stateStore) => {
    const forgersListStr = await stateStore.consensus.get(constants_1.CONSENSUS_STATE_FORGERS_LIST_KEY);
    if (!forgersListStr) {
        return [];
    }
    return JSON.parse(forgersListStr);
};
const _setForgersList = (stateStore, forgersList) => {
    const forgersListStr = JSON.stringify(forgersList);
    stateStore.consensus.set(constants_1.CONSENSUS_STATE_FORGERS_LIST_KEY, forgersListStr);
};
exports.getVoteWeights = async (stateStore) => {
    const voteWeightsStr = await stateStore.consensus.get(constants_1.CONSENSUS_STATE_VOTE_WEIGHTS_KEY);
    if (!voteWeightsStr) {
        return [];
    }
    return JSON.parse(voteWeightsStr);
};
const _setVoteWeights = (stateStore, voteWeights) => {
    const voteWeightsStr = JSON.stringify(voteWeights);
    stateStore.consensus.set(constants_1.CONSENSUS_STATE_VOTE_WEIGHTS_KEY, voteWeightsStr);
};
exports.deleteForgersListUntilRound = async (round, stateStore) => {
    debug('Deleting list until round: ', round);
    const forgersList = await exports.getForgersList(stateStore);
    const newForgersList = forgersList.filter(fl => fl.round >= round);
    _setForgersList(stateStore, newForgersList);
};
exports.deleteForgersListAfterRound = async (round, stateStore) => {
    debug('Deleting list after round: ', round);
    const forgersList = await exports.getForgersList(stateStore);
    const newForgersList = forgersList.filter(fl => fl.round <= round);
    _setForgersList(stateStore, newForgersList);
};
exports.deleteVoteWeightsUntilRound = async (round, stateStore) => {
    debug('Deleting voteWeights until round: ', round);
    const voteWeights = await exports.getVoteWeights(stateStore);
    const newVoteWeights = voteWeights.filter(vw => vw.round >= round);
    _setVoteWeights(stateStore, newVoteWeights);
};
exports.deleteVoteWeightsAfterRound = async (round, stateStore) => {
    debug('Deleting voteWeights after round: ', round);
    const voteWeights = await exports.getVoteWeights(stateStore);
    const newVoteWeights = voteWeights.filter(vw => vw.round <= round);
    _setVoteWeights(stateStore, newVoteWeights);
};
exports.shuffleDelegateList = (previousRoundSeed1, addresses) => {
    const delegateList = [...addresses].map(delegate => ({
        address: delegate,
    }));
    for (const delegate of delegateList) {
        const addressBuffer = lisk_cryptography_1.intToBuffer(delegate.address.slice(0, -1), SIZE_UINT64);
        const seedSource = Buffer.concat([previousRoundSeed1, addressBuffer]);
        delegate.roundHash = lisk_cryptography_1.hash(seedSource);
    }
    delegateList.sort((delegate1, delegate2) => {
        const diff = delegate1.roundHash.compare(delegate2.roundHash);
        if (diff !== 0) {
            return diff;
        }
        return delegate1.address.localeCompare(delegate2.address, 'en');
    });
    return delegateList.map(delegate => delegate.address);
};
exports.getForgerAddressesForRound = async (round, stateStore) => {
    var _a;
    const forgersList = await exports.getForgersList(stateStore);
    const delegateAddresses = (_a = forgersList.find(fl => fl.round === round)) === null || _a === void 0 ? void 0 : _a.delegates;
    if (!delegateAddresses) {
        throw new Error(`No delegate list found for round: ${round}`);
    }
    return delegateAddresses;
};
exports.isCurrentlyPunished = (height, pomHeights) => {
    if (pomHeights.length === 0) {
        return false;
    }
    const lastPomHeight = Math.max(...pomHeights);
    if (height - lastPomHeight < constants_1.PUNISHMENT_PERIOD) {
        return true;
    }
    return false;
};
const _getTotalVoteWeight = (delegateWeights) => delegateWeights.reduce((prev, current) => prev + BigInt(current.voteWeight), BigInt(0));
const _pickStandByDelegate = (delegateWeights, randomSeed) => {
    const seedNumber = randomSeed.readBigUInt64BE();
    const totalVoteWeight = _getTotalVoteWeight(delegateWeights);
    let threshold = seedNumber % totalVoteWeight;
    for (let i = 0; i < delegateWeights.length; i += 1) {
        const voteWeight = BigInt(delegateWeights[i].voteWeight);
        if (voteWeight > threshold) {
            return i;
        }
        threshold -= voteWeight;
    }
    return -1;
};
class DelegatesList {
    constructor({ activeDelegates, standbyDelegates, standbyThreshold, voteWeightCapRate, rounds, chain, }) {
        this.activeDelegates = activeDelegates;
        this.standbyDelegates = standbyDelegates;
        this.standbyThreshold = standbyThreshold;
        this.voteWeightCapRate = voteWeightCapRate;
        this.rounds = rounds;
        this.chain = chain;
    }
    async createVoteWeightsSnapshot(height, stateStore, roundOffset = constants_1.DEFAULT_ROUND_OFFSET) {
        var _a;
        const round = this.rounds.calcRound(height) + roundOffset;
        debug(`Creating vote weight snapshot for round: ${round}`);
        const originalDelegates = await this.chain.dataAccess.getDelegates();
        const updatedAccounts = stateStore.account.getUpdated();
        const updatedAccountsMap = {};
        for (const account of updatedAccounts) {
            if (account.username) {
                updatedAccountsMap[account.address] = account;
            }
        }
        for (const delegate of originalDelegates) {
            if (updatedAccountsMap[delegate.address] === undefined) {
                updatedAccountsMap[delegate.address] = delegate;
            }
        }
        const delegates = [...Object.values(updatedAccountsMap)];
        for (const account of delegates) {
            if (exports.isCurrentlyPunished(height, account.delegate.pomHeights)) {
                account.totalVotesReceived = BigInt(0);
                continue;
            }
            const selfVote = account.votes.find(vote => vote.delegateAddress === account.address);
            const cappedValue = ((_a = selfVote === null || selfVote === void 0 ? void 0 : selfVote.amount) !== null && _a !== void 0 ? _a : BigInt(0)) * BigInt(this.voteWeightCapRate);
            if (account.totalVotesReceived > cappedValue) {
                account.totalVotesReceived = cappedValue;
            }
        }
        delegates.sort((a, b) => {
            const diff = b.totalVotesReceived - a.totalVotesReceived;
            if (diff > BigInt(0)) {
                return 1;
            }
            if (diff < BigInt(0)) {
                return -1;
            }
            return a.address.localeCompare(b.address, 'en');
        });
        const activeDelegates = [];
        const standbyDelegates = [];
        for (const account of delegates) {
            if (account.delegate.isBanned) {
                continue;
            }
            if (activeDelegates.length < this.activeDelegates) {
                activeDelegates.push({
                    address: account.address,
                    voteWeight: account.totalVotesReceived.toString(),
                });
                continue;
            }
            if (account.totalVotesReceived >= this.standbyThreshold) {
                standbyDelegates.push({
                    address: account.address,
                    voteWeight: account.totalVotesReceived.toString(),
                });
                continue;
            }
            if (standbyDelegates.length < this.standbyDelegates) {
                standbyDelegates.push({
                    address: account.address,
                    voteWeight: account.totalVotesReceived.toString(),
                });
                continue;
            }
            break;
        }
        const delegateVoteWeights = activeDelegates.concat(standbyDelegates);
        const voteWeight = {
            round,
            delegates: delegateVoteWeights,
        };
        const voteWeights = await exports.getVoteWeights(stateStore);
        const voteWeightsIndex = voteWeights.findIndex(vw => vw.round === round);
        if (voteWeightsIndex > -1) {
            voteWeights[voteWeightsIndex] = voteWeight;
        }
        else {
            voteWeights.push(voteWeight);
        }
        _setVoteWeights(stateStore, voteWeights);
    }
    async updateForgersList(round, randomSeed, stateStore) {
        if (!randomSeed.length) {
            throw new Error(`Random seed must be provided`);
        }
        const voteWeights = await exports.getVoteWeights(stateStore);
        const voteWeight = voteWeights.find(vw => vw.round === round);
        if (!voteWeight) {
            throw new Error(`Corresponding vote weight for round ${round} not found`);
        }
        const hasStandbySlot = voteWeight.delegates.length > this.activeDelegates;
        const activeDelegateAddresses = voteWeight.delegates
            .slice(0, this.activeDelegates)
            .map(vw => vw.address);
        const standbyDelegateAddresses = [];
        const standbyDelegateVoteWeights = hasStandbySlot
            ? voteWeight.delegates.slice(this.activeDelegates)
            : [];
        if (standbyDelegateVoteWeights.length > 0 &&
            standbyDelegateVoteWeights.length <= this.standbyDelegates) {
            for (const delegate of standbyDelegateVoteWeights) {
                standbyDelegateAddresses.push(delegate.address);
            }
        }
        else if (standbyDelegateVoteWeights.length > this.standbyDelegates) {
            for (let i = 0; i < this.standbyDelegates; i += 1) {
                const standbyDelegateIndex = _pickStandByDelegate(standbyDelegateVoteWeights, randomSeed[i % randomSeed.length]);
                if (standbyDelegateIndex < 0) {
                    throw new Error('Fail to pick standby delegate');
                }
                standbyDelegateAddresses.push(standbyDelegateVoteWeights[standbyDelegateIndex].address);
                standbyDelegateVoteWeights.splice(standbyDelegateIndex, 1);
            }
        }
        const delegates = activeDelegateAddresses.concat(standbyDelegateAddresses);
        const shuffledDelegates = exports.shuffleDelegateList(randomSeed[0], delegates);
        const forgerList = {
            round,
            delegates: shuffledDelegates,
            standby: standbyDelegateAddresses,
        };
        const forgersList = await exports.getForgersList(stateStore);
        const existingIndex = forgersList.findIndex(fl => fl.round === round);
        if (existingIndex > -1) {
            forgersList[existingIndex] = forgerList;
        }
        else {
            forgersList.push(forgerList);
        }
        _setForgersList(stateStore, forgersList);
    }
    async getDelegateList(round) {
        var _a;
        const forgersListStr = await this.chain.dataAccess.getConsensusState(constants_1.CONSENSUS_STATE_FORGERS_LIST_KEY);
        const forgersList = forgersListStr !== undefined
            ? JSON.parse(forgersListStr)
            : [];
        const delegateAddresses = (_a = forgersList.find(fl => fl.round === round)) === null || _a === void 0 ? void 0 : _a.delegates;
        if (!delegateAddresses) {
            throw new Error(`No delegate list found for round: ${round}`);
        }
        return delegateAddresses;
    }
    async verifyBlockForger(block) {
        const currentSlot = this.chain.slots.getSlotNumber(block.timestamp);
        const currentRound = this.rounds.calcRound(block.height);
        const delegateList = await this.getDelegateList(currentRound);
        if (!delegateList.length) {
            throw new Error(`Failed to verify slot: ${currentSlot} for block ID: ${block.id} - No delegateList was found`);
        }
        const expectedForgerAddress = delegateList[currentSlot % delegateList.length];
        if (!expectedForgerAddress ||
            lisk_cryptography_1.getAddressFromPublicKey(block.generatorPublicKey) !==
                expectedForgerAddress) {
            throw new Error(`Failed to verify slot: ${currentSlot}. Block ID: ${block.id}. Block Height: ${block.height}`);
        }
        return true;
    }
}
exports.DelegatesList = DelegatesList;
//# sourceMappingURL=delegates_list.js.map