"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const Debug = require("debug");
const events_1 = require("events");
const errors_1 = require("./errors");
const job_1 = require("./job");
const min_heap_1 = require("./min_heap");
const transaction_list_1 = require("./transaction_list");
const types_1 = require("./types");
const debug = Debug('lisk:transaction_pool');
exports.DEFAULT_MAX_TRANSACTIONS = 4096;
exports.DEFAULT_MAX_TRANSACTIONS_PER_ACCOUNT = 64;
exports.DEFAULT_MIN_ENTRANCE_FEE_PRIORITY = BigInt(0);
exports.DEFAULT_EXPIRY_TIME = 3 * 60 * 60 * 1000;
exports.DEFAULT_EXPIRE_INTERVAL = 60 * 60 * 1000;
exports.DEFAULT_MINIMUM_REPLACEMENT_FEE_DIFFERENCE = BigInt(10);
exports.DEFAULT_REORGANIZE_TIME = 500;
exports.events = {
    EVENT_TRANSACTION_REMOVED: 'EVENT_TRANSACTION_REMOVED',
};
class TransactionPool {
    constructor(config) {
        var _a, _b, _c, _d, _e, _f;
        this.events = new events_1.EventEmitter();
        this._feePriorityQueue = new min_heap_1.MinHeap();
        this._allTransactions = {};
        this._transactionList = {};
        this._applyFunction = config.applyTransactions;
        this._maxTransactions = (_a = config.maxTransactions) !== null && _a !== void 0 ? _a : exports.DEFAULT_MAX_TRANSACTIONS;
        this._maxTransactionsPerAccount = (_b = config.maxTransactionsPerAccount) !== null && _b !== void 0 ? _b : exports.DEFAULT_MAX_TRANSACTIONS_PER_ACCOUNT;
        this._transactionExpiryTime = (_c = config.transactionExpiryTime) !== null && _c !== void 0 ? _c : exports.DEFAULT_EXPIRY_TIME;
        this._minEntranceFeePriority = (_d = config.minEntranceFeePriority) !== null && _d !== void 0 ? _d : exports.DEFAULT_MIN_ENTRANCE_FEE_PRIORITY;
        this._transactionReorganizationInterval = (_e = config.transactionReorganizationInterval) !== null && _e !== void 0 ? _e : exports.DEFAULT_REORGANIZE_TIME;
        this._minReplacementFeeDifference = (_f = config.minReplacementFeeDifference) !== null && _f !== void 0 ? _f : exports.DEFAULT_MINIMUM_REPLACEMENT_FEE_DIFFERENCE;
        this._reorganizeJob = new job_1.Job(() => this._reorganize(), this._transactionReorganizationInterval);
        this._expireJob = new job_1.Job(() => this._expire(), exports.DEFAULT_EXPIRE_INTERVAL);
    }
    async start() {
        this._reorganizeJob.start();
        this._expireJob.start();
    }
    stop() {
        this._reorganizeJob.stop();
        this._expireJob.stop();
    }
    getAll() {
        return Object.values(this._allTransactions);
    }
    get(id) {
        return this._allTransactions[id];
    }
    contains(id) {
        return this._allTransactions[id] !== undefined;
    }
    async add(incomingTx) {
        if (this._allTransactions[incomingTx.id]) {
            debug('Received duplicate transaction', incomingTx.id);
            return { status: types_1.Status.OK, errors: [] };
        }
        incomingTx.feePriority = this._calculateFeePriority(incomingTx);
        if (incomingTx.feePriority < this._minEntranceFeePriority) {
            const error = new errors_1.TransactionPoolError(`Rejecting transaction due to failed minimum entrance fee priority requirement`, incomingTx.id, '.fee', incomingTx.feePriority.toString(), this._minEntranceFeePriority.toString());
            return { status: types_1.Status.FAIL, errors: [error] };
        }
        const lowestFeePriorityTrx = this._feePriorityQueue.peek();
        if (Object.keys(this._allTransactions).length >= this._maxTransactions &&
            lowestFeePriorityTrx &&
            incomingTx.feePriority <= lowestFeePriorityTrx.key) {
            const error = new errors_1.TransactionPoolError(`Rejecting transaction due to fee priority when the pool is full`, incomingTx.id, '.fee', incomingTx.feePriority.toString(), lowestFeePriorityTrx.key.toString());
            return { status: types_1.Status.FAIL, errors: [error] };
        }
        const incomingTxAddress = lisk_cryptography_1.getAddressFromPublicKey(incomingTx.senderPublicKey);
        const transactionsResponses = await this._applyFunction([incomingTx]);
        const txStatus = this._getStatus(transactionsResponses[0]);
        if (txStatus === types_1.TransactionStatus.INVALID) {
            return { status: types_1.Status.FAIL, errors: transactionsResponses[0].errors };
        }
        const exceededTransactionsCount = Object.keys(this._allTransactions).length - this._maxTransactions;
        if (exceededTransactionsCount >= 0) {
            const isEvicted = this._evictUnprocessable();
            if (!isEvicted) {
                this._evictProcessable();
            }
        }
        if (!this._transactionList[incomingTxAddress]) {
            this._transactionList[incomingTxAddress] = new transaction_list_1.TransactionList(incomingTxAddress, {
                maxSize: this._maxTransactionsPerAccount,
                minReplacementFeeDifference: this._minReplacementFeeDifference,
            });
        }
        const { added, removedID, reason } = this._transactionList[incomingTxAddress].add(incomingTx, txStatus === types_1.TransactionStatus.PROCESSABLE);
        if (!added) {
            return {
                status: types_1.Status.FAIL,
                errors: [new errors_1.TransactionPoolError(reason, incomingTx.id)],
            };
        }
        if (removedID) {
            debug('Removing from transaction pool with id', removedID);
            const removedTx = this._allTransactions[removedID];
            delete this._allTransactions[removedID];
            this.events.emit(exports.events.EVENT_TRANSACTION_REMOVED, {
                id: removedTx.id,
                nonce: removedTx.nonce.toString(),
                senderPublicKey: removedTx.senderPublicKey,
                reason: 'Transaction List executed remove',
            });
        }
        incomingTx.receivedAt = new Date();
        this._allTransactions[incomingTx.id] = incomingTx;
        this._feePriorityQueue.push(this._calculateFeePriority(incomingTx), incomingTx.id);
        return { status: types_1.Status.OK, errors: [] };
    }
    remove(tx) {
        var _a;
        const foundTx = this._allTransactions[tx.id];
        if (!foundTx) {
            return false;
        }
        delete this._allTransactions[tx.id];
        debug('Removing from transaction pool with id', tx.id);
        const senderId = lisk_cryptography_1.getAddressFromPublicKey(foundTx.senderPublicKey);
        this._transactionList[senderId].remove(tx.nonce);
        if (this._transactionList[senderId].size === 0) {
            delete this._transactionList[senderId];
        }
        this._feePriorityQueue.clear();
        for (const txObject of this.getAll()) {
            this._feePriorityQueue.push((_a = txObject.feePriority) !== null && _a !== void 0 ? _a : this._calculateFeePriority(txObject), txObject.id);
        }
        return true;
    }
    getProcessableTransactions() {
        const processableTransactions = {};
        for (const address of Object.keys(this._transactionList)) {
            const transactions = this._transactionList[address].getProcessable();
            if (transactions.length !== 0) {
                processableTransactions[address] = [...transactions];
            }
        }
        return processableTransactions;
    }
    _calculateFeePriority(trx) {
        return (trx.fee - trx.minFee) / BigInt(trx.getBytes().length);
    }
    _getStatus(txResponse) {
        if (txResponse.status === types_1.Status.OK) {
            debug('Received PROCESSABLE transaction');
            return types_1.TransactionStatus.PROCESSABLE;
        }
        const txResponseErrors = txResponse.errors;
        if (txResponse.errors.length === 1 &&
            txResponseErrors[0].dataPath === '.nonce' &&
            txResponseErrors[0].actual &&
            txResponseErrors[0].expected &&
            BigInt(txResponseErrors[0].actual) > BigInt(txResponseErrors[0].expected)) {
            debug('Received UNPROCESSABLE transaction');
            return types_1.TransactionStatus.UNPROCESSABLE;
        }
        debug('Received INVALID transaction');
        return types_1.TransactionStatus.INVALID;
    }
    _evictUnprocessable() {
        const unprocessableFeePriorityHeap = new min_heap_1.MinHeap();
        for (const txList of Object.values(this._transactionList)) {
            const unprocessableTransactions = txList.getUnprocessable();
            for (const unprocessableTx of unprocessableTransactions) {
                unprocessableFeePriorityHeap.push(unprocessableTx.feePriority, unprocessableTx);
            }
        }
        if (unprocessableFeePriorityHeap.count < 1) {
            return false;
        }
        const evictedTransaction = unprocessableFeePriorityHeap.pop();
        if (!evictedTransaction) {
            return false;
        }
        this.events.emit(exports.events.EVENT_TRANSACTION_REMOVED, {
            id: evictedTransaction.value.id,
            nonce: evictedTransaction.value.nonce.toString(),
            senderPublicKey: evictedTransaction.value.senderPublicKey,
            reason: 'Pool exceeded the size limit',
        });
        return this.remove(evictedTransaction.value);
    }
    _evictProcessable() {
        const processableFeePriorityHeap = new min_heap_1.MinHeap();
        for (const txList of Object.values(this._transactionList)) {
            const processableTransactions = txList.getProcessable();
            if (processableTransactions.length) {
                const processableTransactionWithHighestNonce = processableTransactions[processableTransactions.length - 1];
                processableFeePriorityHeap.push(processableTransactionWithHighestNonce.feePriority, processableTransactionWithHighestNonce);
            }
        }
        if (processableFeePriorityHeap.count < 1) {
            return false;
        }
        const evictedTransaction = processableFeePriorityHeap.pop();
        if (!evictedTransaction) {
            return false;
        }
        this.events.emit(exports.events.EVENT_TRANSACTION_REMOVED, {
            id: evictedTransaction.value.id,
            nonce: evictedTransaction.value.nonce.toString(),
            senderPublicKey: evictedTransaction.value.senderPublicKey,
            reason: 'Pool exceeded the size limit',
        });
        return this.remove(evictedTransaction.value);
    }
    async _reorganize() {
        for (const txList of Object.values(this._transactionList)) {
            const promotableTransactions = txList.getPromotable();
            if (!promotableTransactions.length) {
                continue;
            }
            const processableTransactions = txList.getProcessable();
            const allTransactions = [
                ...processableTransactions,
                ...promotableTransactions,
            ];
            const applyResults = await this._applyFunction(allTransactions);
            const successfulTransactionIds = [];
            let nonProcessableIds = [];
            let firstInvalidTransactionId;
            for (const result of applyResults) {
                const txApplyStatus = this._getStatus(result);
                if (txApplyStatus === types_1.TransactionStatus.INVALID) {
                    firstInvalidTransactionId = result.id;
                    break;
                }
                if (txApplyStatus === types_1.TransactionStatus.UNPROCESSABLE) {
                    nonProcessableIds.push(result.id);
                }
                successfulTransactionIds.push(result.id);
            }
            const trxsToPromote = successfulTransactionIds.filter(id => !nonProcessableIds.includes(id));
            txList.promote(promotableTransactions.filter(tx => trxsToPromote.includes(tx.id)));
            const invalidTransaction = firstInvalidTransactionId
                ? allTransactions.find(tx => tx.id == firstInvalidTransactionId)
                : undefined;
            if (invalidTransaction) {
                for (const tx of allTransactions) {
                    if (tx.nonce >= invalidTransaction.nonce) {
                        this.events.emit(exports.events.EVENT_TRANSACTION_REMOVED, {
                            id: tx.id,
                            nonce: tx.nonce.toString(),
                            senderPublicKey: tx.senderPublicKey,
                            reason: `Invalid transaction ${invalidTransaction.id}`,
                        });
                        this.remove(tx);
                    }
                }
            }
        }
    }
    async _expire() {
        for (const transaction of Object.values(this._allTransactions)) {
            const timeDifference = Math.round(Math.abs(transaction.receivedAt.getTime() - new Date().getTime()));
            if (timeDifference > this._transactionExpiryTime) {
                this.events.emit(exports.events.EVENT_TRANSACTION_REMOVED, {
                    id: transaction.id,
                    nonce: transaction.nonce.toString(),
                    senderPublicKey: transaction.senderPublicKey,
                    reason: 'Transaction exceeded the expiry time',
                });
                this.remove(transaction);
            }
        }
    }
}
exports.TransactionPool = TransactionPool;
//# sourceMappingURL=transaction_pool.js.map