"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const base_transaction_1 = require("./base_transaction");
const errors_1 = require("./errors");
const response_1 = require("./response");
const utils_1 = require("./utils");
exports.multisignatureAssetFormatSchema = {
    type: 'object',
    required: ['mandatoryKeys', 'optionalKeys', 'numberOfSignatures'],
    properties: {
        numberOfSignatures: {
            type: 'integer',
            minimum: 1,
            maximum: 64,
        },
        optionalKeys: {
            type: 'array',
            uniqueItems: true,
            minItems: 0,
            maxItems: 64,
            items: {
                type: 'string',
                format: 'publicKey',
            },
        },
        mandatoryKeys: {
            type: 'array',
            uniqueItems: true,
            minItems: 0,
            maxItems: 64,
            items: {
                type: 'string',
                format: 'publicKey',
            },
        },
    },
};
const setMemberAccounts = async (store, membersPublicKeys) => {
    for (const memberPublicKey of membersPublicKeys) {
        const address = lisk_cryptography_1.getAddressFromPublicKey(memberPublicKey);
        const memberAccount = await store.account.getOrDefault(address);
        memberAccount.publicKey = memberAccount.publicKey || memberPublicKey;
        store.account.set(memberAccount.address, memberAccount);
    }
};
class MultisignatureTransaction extends base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        this.MAX_KEYS_COUNT = 64;
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        this.asset = (tx.asset || {});
    }
    assetToBytes() {
        const { mandatoryKeys, optionalKeys, numberOfSignatures } = this.asset;
        const mandatoryKeysBuffer = Buffer.from(mandatoryKeys.join(''), 'hex');
        const optionalKeysBuffer = Buffer.from(optionalKeys.join(''), 'hex');
        const assetBuffer = Buffer.concat([
            lisk_cryptography_1.intToBuffer(mandatoryKeys.length, 1),
            mandatoryKeysBuffer,
            lisk_cryptography_1.intToBuffer(optionalKeys.length, 1),
            optionalKeysBuffer,
            lisk_cryptography_1.intToBuffer(numberOfSignatures, 1),
        ]);
        return assetBuffer;
    }
    async prepare(store) {
        const membersAddresses = [
            ...this.asset.mandatoryKeys,
            ...this.asset.optionalKeys,
        ].map(publicKey => ({ address: lisk_cryptography_1.getAddressFromPublicKey(publicKey) }));
        await store.account.cache([
            {
                address: this.senderId,
            },
            ...membersAddresses,
        ]);
    }
    verifyAgainstTransactions(transactions) {
        const errors = transactions
            .filter(tx => tx.type === this.type && tx.senderPublicKey === this.senderPublicKey)
            .map(tx => new errors_1.TransactionError('Register multisignature only allowed once per account.', tx.id, '.asset.multisignature'));
        return errors;
    }
    validateAsset() {
        const schemaErrors = lisk_validator_1.validator.validate(exports.multisignatureAssetFormatSchema, this.asset);
        const errors = errors_1.convertToAssetError(this.id, schemaErrors);
        if (errors.length > 0) {
            return errors;
        }
        const { mandatoryKeys, optionalKeys, numberOfSignatures } = this.asset;
        if (mandatoryKeys.length + optionalKeys.length < numberOfSignatures) {
            errors.push(new errors_1.TransactionError('The numberOfSignatures is bigger than the count of Mandatory and Optional keys', this.id, '.asset.numberOfSignatures', this.asset.numberOfSignatures));
        }
        if (mandatoryKeys.length + optionalKeys.length > this.MAX_KEYS_COUNT ||
            mandatoryKeys.length + optionalKeys.length <= 0) {
            errors.push(new errors_1.TransactionError('The count of Mandatory and Optional keys should be between 1 and 64', this.id, '.asset.optionalKeys .asset.mandatoryKeys', this.asset.numberOfSignatures));
        }
        if (mandatoryKeys.length > numberOfSignatures) {
            errors.push(new errors_1.TransactionError('The numberOfSignatures needs to be equal or bigger than the number of Mandatory keys', this.id, '.asset.numberOfSignatures', this.asset.numberOfSignatures));
        }
        if (errors.length > 0) {
            return errors;
        }
        const repeatedKeys = mandatoryKeys.filter(value => optionalKeys.includes(value));
        if (repeatedKeys.length > 0) {
            errors.push(new errors_1.TransactionError('Invalid combination of Mandatory and Optional keys', this.id, '.asset.mandatoryKeys, .asset.optionalKeys', repeatedKeys.join(', ')));
        }
        if (errors.length > 0) {
            return errors;
        }
        if (mandatoryKeys.length + optionalKeys.length + 1 !==
            this.signatures.length) {
            return [
                new errors_1.TransactionError('The number of mandatory, optional and sender keys should match the number of signatures', this.id),
            ];
        }
        const sortedMandatoryKeys = [...mandatoryKeys].sort();
        const sortedOptionalKeys = [...optionalKeys].sort();
        for (let i = 0; i < sortedMandatoryKeys.length; i += 1) {
            if (mandatoryKeys[i] !== sortedMandatoryKeys[i]) {
                errors.push(new errors_1.TransactionError('Mandatory keys should be sorted lexicographically', this.id, '.asset.mandatoryKeys', mandatoryKeys.join(', ')));
                break;
            }
        }
        for (let i = 0; i < sortedOptionalKeys.length; i += 1) {
            if (optionalKeys[i] !== sortedOptionalKeys[i]) {
                errors.push(new errors_1.TransactionError('Optional keys should be sorted lexicographically', this.id, '.asset.optionalKeys', optionalKeys.join(', ')));
                break;
            }
        }
        return errors;
    }
    async applyAsset(store) {
        const errors = [];
        const sender = await store.account.get(this.senderId);
        if (sender.keys.numberOfSignatures > 0) {
            errors.push(new errors_1.TransactionError('Register multisignature only allowed once per account.', this.id, '.signatures'));
        }
        sender.keys = {
            numberOfSignatures: this.asset.numberOfSignatures,
            mandatoryKeys: this.asset.mandatoryKeys,
            optionalKeys: this.asset.optionalKeys,
        };
        store.account.set(sender.address, sender);
        await setMemberAccounts(store, sender.keys.mandatoryKeys);
        await setMemberAccounts(store, sender.keys.optionalKeys);
        return errors;
    }
    async undoAsset(store) {
        const sender = await store.account.get(this.senderId);
        sender.keys = {
            mandatoryKeys: [],
            optionalKeys: [],
            numberOfSignatures: 0,
        };
        store.account.set(sender.address, sender);
        return [];
    }
    async verifySignatures(store) {
        const { networkIdentifier } = store.chain;
        const transactionBytes = this.getBasicBytes();
        const networkIdentifierBytes = lisk_cryptography_1.hexToBuffer(networkIdentifier);
        const transactionWithNetworkIdentifierBytes = Buffer.concat([
            networkIdentifierBytes,
            transactionBytes,
        ]);
        const { mandatoryKeys, optionalKeys } = this.asset;
        if (mandatoryKeys.length + optionalKeys.length + 1 !==
            this.signatures.length) {
            return response_1.createResponse(this.id, [
                new errors_1.TransactionError('There are missing signatures'),
            ]);
        }
        if (this.signatures.includes('')) {
            return response_1.createResponse(this.id, [
                new errors_1.TransactionError('A signature is required for each registered key.'),
            ]);
        }
        const { valid, error } = utils_1.validateSignature(this.senderPublicKey, this.signatures[0], transactionWithNetworkIdentifierBytes);
        if (!valid) {
            return response_1.createResponse(this.id, [error]);
        }
        const mandatorySignaturesErrors = utils_1.validateKeysSignatures(mandatoryKeys, this.signatures.slice(1, mandatoryKeys.length + 1), transactionWithNetworkIdentifierBytes);
        if (mandatorySignaturesErrors.length) {
            return response_1.createResponse(this.id, mandatorySignaturesErrors);
        }
        const optionalSignaturesErrors = utils_1.validateKeysSignatures(optionalKeys, this.signatures.slice(mandatoryKeys.length + 1), transactionWithNetworkIdentifierBytes);
        if (optionalSignaturesErrors.length) {
            return response_1.createResponse(this.id, optionalSignaturesErrors);
        }
        return response_1.createResponse(this.id, []);
    }
    sign(networkIdentifier, senderPassphrase, passphrases, keys) {
        utils_1.sortKeysAscending(this.asset.mandatoryKeys);
        utils_1.sortKeysAscending(this.asset.optionalKeys);
        const { publicKey } = lisk_cryptography_1.getAddressAndPublicKeyFromPassphrase(senderPassphrase);
        if (this.senderPublicKey !== '' && this.senderPublicKey !== publicKey) {
            throw new Error('Transaction senderPublicKey does not match public key from passphrase');
        }
        this.senderPublicKey = publicKey;
        const networkIdentifierBytes = lisk_cryptography_1.hexToBuffer(networkIdentifier);
        const transactionWithNetworkIdentifierBytes = Buffer.concat([
            networkIdentifierBytes,
            this.getBasicBytes(),
        ]);
        this.signatures.push(lisk_cryptography_1.signData(lisk_cryptography_1.hash(transactionWithNetworkIdentifierBytes), senderPassphrase));
        if (keys && passphrases) {
            const keysAndPassphrases = utils_1.buildPublicKeyPassphraseDict([...passphrases]);
            utils_1.sortKeysAscending(keys.mandatoryKeys);
            utils_1.sortKeysAscending(keys.optionalKeys);
            for (const aKey of [...keys.mandatoryKeys, ...keys.optionalKeys]) {
                if (keysAndPassphrases[aKey]) {
                    const { passphrase } = keysAndPassphrases[aKey];
                    this.signatures.push(lisk_cryptography_1.signData(lisk_cryptography_1.hash(transactionWithNetworkIdentifierBytes), passphrase));
                }
                else {
                    this.signatures.push('');
                }
            }
        }
        this._id = utils_1.getId(this.getBytes());
    }
}
exports.MultisignatureTransaction = MultisignatureTransaction;
MultisignatureTransaction.TYPE = 12;
//# sourceMappingURL=12_multisignature_transaction.js.map