"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const cryptography = require("@liskhq/lisk-cryptography");
const _10_delegate_transaction_1 = require("./10_delegate_transaction");
const _12_multisignature_transaction_1 = require("./12_multisignature_transaction");
const _13_vote_transaction_1 = require("./13_vote_transaction");
const _14_unlock_transaction_1 = require("./14_unlock_transaction");
const _8_transfer_transaction_1 = require("./8_transfer_transaction");
const utils_1 = require("./utils");
const transactionMap = {
    8: _8_transfer_transaction_1.TransferTransaction,
    10: _10_delegate_transaction_1.DelegateTransaction,
    12: _12_multisignature_transaction_1.MultisignatureTransaction,
    13: _13_vote_transaction_1.VoteTransaction,
    14: _14_unlock_transaction_1.UnlockTransaction,
};
const sanitizeSignaturesArray = (tx, keys) => {
    let numberOfSignatures = keys.mandatoryKeys.length + keys.optionalKeys.length;
    if (tx.type === _12_multisignature_transaction_1.MultisignatureTransaction.TYPE) {
        numberOfSignatures += 1;
    }
    for (let i = 0; i < numberOfSignatures; i += 1) {
        if (tx.signatures[i] === undefined) {
            tx.signatures[i] = '';
        }
    }
};
exports.signMultiSignatureTransaction = (options) => {
    const { transaction, passphrase, networkIdentifier, keys } = options;
    if (transaction.type === undefined || transaction.type === null) {
        throw new Error('Transaction type is required.');
    }
    if (!Object.keys(transactionMap).includes(String(transaction.type))) {
        throw new Error('Invalid transaction type.');
    }
    utils_1.sortKeysAscending(keys.mandatoryKeys);
    utils_1.sortKeysAscending(keys.optionalKeys);
    const TransactionClass = transactionMap[transaction.type];
    const tx = new TransactionClass({
        ...transaction,
        networkIdentifier,
    });
    const { publicKey } = cryptography.getPrivateAndPublicKeyFromPassphrase(passphrase);
    const networkIdentifierBytes = Buffer.from(networkIdentifier, 'hex');
    const transactionWithNetworkIdentifierBytes = Buffer.concat([
        networkIdentifierBytes,
        tx.getBasicBytes(),
    ]);
    const signature = cryptography.signData(cryptography.hash(transactionWithNetworkIdentifierBytes), passphrase);
    if (tx.signatures.includes(signature)) {
        sanitizeSignaturesArray(tx, keys);
        return tx;
    }
    const mandatoryKeyIndex = keys.mandatoryKeys.findIndex(aPublicKey => aPublicKey === publicKey);
    const optionalKeyIndex = keys.optionalKeys.findIndex(aPublicKey => aPublicKey === publicKey);
    if (mandatoryKeyIndex !== -1) {
        let signatureOffset = 0;
        if (tx.type === _12_multisignature_transaction_1.MultisignatureTransaction.TYPE) {
            signatureOffset = 1;
        }
        tx.signatures[mandatoryKeyIndex + signatureOffset] = signature;
    }
    if (optionalKeyIndex !== -1) {
        let signatureOffset = 0;
        if (tx.type === _12_multisignature_transaction_1.MultisignatureTransaction.TYPE) {
            signatureOffset = 1;
        }
        tx.signatures[keys.mandatoryKeys.length + optionalKeyIndex + signatureOffset] = signature;
    }
    sanitizeSignaturesArray(tx, keys);
    return tx;
};
//# sourceMappingURL=sign_multi_signature_transaction.js.map