"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("../errors");
const format_1 = require("./format");
const sign_and_validate_1 = require("./sign_and_validate");
exports.verifySenderPublicKey = (id, sender, publicKey) => sender.publicKey && sender.publicKey !== publicKey
    ? new errors_1.TransactionError('Invalid sender publicKey', id, '.senderPublicKey', publicKey, sender.publicKey)
    : undefined;
exports.verifyMinRemainingBalance = (id, account, minRemainingBalance) => {
    if (account.balance < minRemainingBalance) {
        return new errors_1.TransactionError(`Account does not have enough minimum remaining LSK: ${account.address}, balance: ${format_1.convertBeddowsToLSK(account.balance.toString())}`, id, '.balance', account.balance.toString(), minRemainingBalance.toString());
    }
    return undefined;
};
exports.verifyAccountNonce = (id, account, nonce) => {
    if (nonce < account.nonce) {
        return new errors_1.TransactionError(`Incompatible transaction nonce for account: ${account.address}, Tx Nonce: ${nonce.toString()}, Account Nonce: ${account.nonce.toString()}`, id, '.nonce', nonce.toString(), account.nonce.toString());
    }
    if (nonce > account.nonce) {
        return new errors_1.TransactionError(`Transaction nonce for account: ${account.address} is higher than expected, Tx Nonce: ${nonce.toString()}, Account Nonce: ${account.nonce.toString()}`, id, '.nonce', nonce.toString(), account.nonce.toString());
    }
    return undefined;
};
exports.isMultisignatureAccount = (account) => !!((account.keys.mandatoryKeys.length > 0 ||
    account.keys.optionalKeys.length > 0) &&
    account.keys.numberOfSignatures);
exports.validateKeysSignatures = (keys, signatures, transactionBytes) => {
    const errors = [];
    for (let i = 0; i < keys.length; i += 1) {
        if (signatures[i].length === 0) {
            errors.push(new errors_1.TransactionError('Invalid signatures format. signatures should not include empty string.', undefined, '.signatures', signatures.join(',')));
            break;
        }
        const { error } = sign_and_validate_1.validateSignature(keys[i], signatures[i], transactionBytes);
        if (error) {
            errors.push(error);
        }
    }
    return errors;
};
exports.verifyMultiSignatureTransaction = (id, sender, signatures, transactionBytes) => {
    const errors = [];
    const { mandatoryKeys, optionalKeys, numberOfSignatures } = sender.keys;
    const numMandatoryKeys = mandatoryKeys.length;
    const numOptionalKeys = optionalKeys.length;
    const nonEmptySignaturesCount = signatures.filter(k => k.length !== 0).length;
    if (nonEmptySignaturesCount !== numberOfSignatures ||
        signatures.length !== numMandatoryKeys + numOptionalKeys) {
        const error = new errors_1.TransactionError(`Transaction signatures does not match required number of signatures: ${numberOfSignatures}`, id, '.signatures', signatures.join(','));
        return [error];
    }
    const mandatoryKeysError = exports.validateKeysSignatures(mandatoryKeys, signatures, transactionBytes);
    errors.push(...mandatoryKeysError);
    for (let k = 0; k < numOptionalKeys; k += 1) {
        const signature = signatures[numMandatoryKeys + k];
        if (signature.length !== 0) {
            const { error } = sign_and_validate_1.validateSignature(optionalKeys[k], signature, transactionBytes);
            if (error) {
                errors.push(error);
            }
        }
    }
    return errors;
};
//# sourceMappingURL=verify.js.map