"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
const processInputs = (nonce, fee, networkIdentifier, username, passphrase) => lisk_transactions_1.registerDelegate({
    nonce,
    fee,
    networkIdentifier,
    passphrase,
    username,
});
class DelegateCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'no-signature': noSignature, }, } = this.parse(DelegateCommand);
        const { nonce, fee, username } = args;
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(fee))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const normalizedFee = lisk_transactions_1.utils.convertLSKToBeddows(fee);
        if (!lisk_validator_1.isValidFee(normalizedFee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        if (noSignature) {
            const noSignatureResult = processInputs(nonce, normalizedFee, networkIdentifier, username);
            this.print(noSignatureResult);
            return;
        }
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = processInputs(nonce, normalizedFee, networkIdentifier, username, passphrase);
        this.print(result);
    }
}
exports.default = DelegateCommand;
DelegateCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
    {
        name: 'username',
        required: true,
        description: 'Username to register as a delegate.',
    },
];
DelegateCommand.description = `
	Creates a transaction which will register the account as a delegate candidate if broadcast to the network.
	`;
DelegateCommand.examples = ['transaction:create:delegate 1 100 lightcurve'];
DelegateCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'no-signature': command_1.flags.boolean(flags_1.flags.noSignature),
};
//# sourceMappingURL=delegate.js.map