"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
class MultisignatureCommand extends base_1.default {
    async run() {
        const { args: { fee: feeSource, nonce }, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'member-passphrase': memberPassphrasesSource, 'optional-key': optionalKeysSource, 'mandatory-key': mandatoryKeysSource, 'number-of-signatures': numberOfSignaturesSource, }, } = this.parse(MultisignatureCommand);
        const numberOfSignatures = numberOfSignaturesSource || 0;
        const mandatoryKeys = mandatoryKeysSource || [];
        const optionalKeys = optionalKeysSource || [];
        const memberPassphrases = memberPassphrasesSource || [];
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(feeSource))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const fee = lisk_transactions_1.utils.convertLSKToBeddows(feeSource);
        if (!lisk_validator_1.isValidFee(fee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const senderPassphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = lisk_transactions_1.registerMultisignature({
            senderPassphrase,
            passphrases: memberPassphrases,
            mandatoryKeys,
            optionalKeys,
            numberOfSignatures,
            networkIdentifier,
            nonce,
            fee,
        });
        this.print(result);
    }
}
exports.default = MultisignatureCommand;
MultisignatureCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
];
MultisignatureCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'number-of-signatures': command_1.flags.integer({
        description: 'Number of signatures required to validate transactions from that account',
    }),
    'mandatory-key': command_1.flags.string({
        multiple: true,
        description: 'Mandatory public key for multisignature account.',
    }),
    'optional-key': command_1.flags.string({
        multiple: true,
        description: 'Optional public key for multisignature account.',
    }),
    'member-passphrase': command_1.flags.string({
        multiple: true,
        description: 'Passphrase for all mandatory and optional keys. Its a less secure way to sign transaction. See command description for details. ',
    }),
};
MultisignatureCommand.description = `
	Creates a transaction which will register the account as a multisignature account if broadcast to the network, using the following arguments:

		1. Mandatory keys to verify signatures for the multisignature account. Signatures for these keys must exists to verify every transaction from that account.
		2. Optional keys to verify signatures for the multisignature account. Signatures for these keys may or may not exists to verify transaction from that account.
		3. Number of signatures required for a transaction from the account to be valid. These value must be minimum of count of mandatory keys.
		4. Member passphrases for every mandatory and optional key provided.

	Use of '--member-passphrase' flag is not encourage for better security practices. This will expose the passphrase to command line and may end up logging into system logs. Use 'transaction:sign' command to sign with member passphrases.
	`;
MultisignatureCommand.examples = [
    'transaction:create:multisignature 1 100 --mandatory-key="xxx" ' +
        '--mandatory-key="yyy" --optional-key="yyy" --optional-key="yyy" ' +
        '--number-of-signatures=4 --passphrase="****" ' +
        '--member-passphrase="****" --member-passphrase="****" ',
];
//# sourceMappingURL=multisignature.js.map