"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
const processInputs = (nonce, fee, networkIdentifier, header1, header2, passphrase) => lisk_transactions_1.reportMisbehavior({
    nonce,
    fee,
    networkIdentifier,
    passphrase,
    header1,
    header2,
});
class PoMCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'no-signature': noSignature, }, } = this.parse(PoMCommand);
        const { nonce, fee, header1: header1Str, header2: header2Str, } = args;
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(fee))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const normalizedFee = lisk_transactions_1.utils.convertLSKToBeddows(fee);
        if (!lisk_validator_1.isValidFee(normalizedFee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        let header1;
        try {
            header1 = JSON.parse(header1Str);
        }
        catch (error) {
            throw new Error(`Invalid block header 1. Fail to parse the input. ${error.message}`);
        }
        let header2;
        try {
            header2 = JSON.parse(header2Str);
        }
        catch (error) {
            throw new Error(`Invalid block header 1. Fail to parse the input. ${error.message}`);
        }
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (noSignature) {
            const noSignatureResult = processInputs(nonce, normalizedFee, networkIdentifier, header1, header2);
            this.print(noSignatureResult);
            return;
        }
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = processInputs(nonce, normalizedFee, networkIdentifier, header1, header2, passphrase);
        this.print(result);
    }
}
exports.default = PoMCommand;
PoMCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
    {
        name: 'header1',
        required: true,
        description: 'Contradicting block header as JSON string.',
    },
    {
        name: 'header2',
        required: true,
        description: 'Contradicting block header as JSON string.',
    },
];
PoMCommand.description = `
	Creates a transaction which will report misbehavior of delegate by providing 2 contradicting block headers.
	`;
PoMCommand.examples = [
    'transaction:create:pom 1 100 "{"height": 3, "version": 2, "maxHeightPrevoted": 30, "blockSignature": "xxx"}" "{"height": 3, "version": 2, "maxHeightPrevoted": 31, "blockSignature": "yyy"}"',
];
PoMCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'no-signature': command_1.flags.boolean(flags_1.flags.noSignature),
};
//# sourceMappingURL=pom.js.map