"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
const createUnlockTransaction = (nonce, fee, networkIdentifier, unlockingObjects, passphrase) => lisk_transactions_1.unlockToken({
    nonce,
    fee,
    networkIdentifier,
    passphrase,
    unlockingObjects,
});
const splitInputs = (unlock) => unlock
    .split(',')
    .filter(Boolean)
    .map(u => u.trim());
const validateUnlocks = (unlocks) => {
    const rawAssetUnlock = [];
    for (const unlock of unlocks) {
        const [delegateAddress, amount, unvoteHeight] = splitInputs(unlock);
        if (!lisk_validator_1.validateAddress(delegateAddress)) {
            throw new error_1.ValidationError('Enter a valid address in LSK string format.');
        }
        const normalizedAmount = lisk_transactions_1.utils.convertLSKToBeddows(amount);
        if (!lisk_validator_1.isValidFee(normalizedAmount)) {
            throw new error_1.ValidationError('Enter a valid amount in number string format.');
        }
        if (!lisk_validator_1.isNumberString(unvoteHeight)) {
            throw new error_1.ValidationError('Enter the unvoteHeight in valid number format.');
        }
        rawAssetUnlock.push({
            delegateAddress,
            amount: normalizedAmount,
            unvoteHeight: parseInt(unvoteHeight, 10),
        });
    }
    return rawAssetUnlock;
};
class UnlockCommand extends base_1.default {
    async run() {
        const { args, flags: { unlock: unlocks, networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'no-signature': noSignature, }, } = this.parse(UnlockCommand);
        const { nonce, fee } = args;
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(fee))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const normalizedFee = lisk_transactions_1.utils.convertLSKToBeddows(fee);
        if (!lisk_validator_1.isValidFee(normalizedFee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        if (!(unlocks === null || unlocks === void 0 ? void 0 : unlocks.length)) {
            throw new error_1.ValidationError('At least one unlock object options must be provided.');
        }
        const unlockingObjects = validateUnlocks(unlocks);
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (noSignature) {
            const noSignatureResult = createUnlockTransaction(nonce, normalizedFee, networkIdentifier, unlockingObjects);
            this.print(noSignatureResult);
            return;
        }
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = createUnlockTransaction(nonce, normalizedFee, networkIdentifier, unlockingObjects, passphrase);
        this.print(result);
    }
}
exports.default = UnlockCommand;
UnlockCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
];
UnlockCommand.description = `
	Creates a transaction which will unlock tokens voted for delegates and add them back to the sender balance.
	`;
UnlockCommand.examples = [
    'transaction:create:unlock 1 100 --unlock="123L,1000000000,500"',
    'transaction:create:unlock 1 100 --unlock="123L,1000000000,500" --unlock="456L,1000000000,500"',
];
UnlockCommand.flags = {
    ...base_1.default.flags,
    unlock: command_1.flags.string({ ...flags_1.flags.unlock, multiple: true }),
    'no-signature': command_1.flags.boolean(flags_1.flags.noSignature),
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
};
//# sourceMappingURL=unlock.js.map