"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../../base"));
const error_1 = require("../../../utils/error");
const flags_1 = require("../../../utils/flags");
const network_identifier_1 = require("../../../utils/network_identifier");
const reader_1 = require("../../../utils/reader");
const processInputs = (nonce, fee, networkIdentifier, votes, passphrase) => lisk_transactions_1.castVotes({
    nonce,
    fee,
    networkIdentifier,
    passphrase,
    votes,
});
const validateAddresses = (inputs) => {
    for (const rawVoteAsset of inputs) {
        lisk_validator_1.validateAddress(rawVoteAsset.delegateAddress);
    }
    return inputs;
};
class VoteCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'no-signature': noSignature, votes, }, } = this.parse(VoteCommand);
        const { nonce, fee } = args;
        if (!lisk_validator_1.isValidNonce(nonce)) {
            throw new error_1.ValidationError('Enter a valid nonce in number string format.');
        }
        if (Number.isNaN(Number(fee))) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const normalizedFee = lisk_transactions_1.utils.convertLSKToBeddows(fee);
        if (!lisk_validator_1.isValidFee(normalizedFee)) {
            throw new error_1.ValidationError('Enter a valid fee in number string format.');
        }
        const targetAddresses = [];
        const votesObjects = votes.map(vote => {
            const voteArr = vote.split(',');
            const [delegateAddress, amount] = voteArr;
            if (!targetAddresses.includes(delegateAddress)) {
                targetAddresses.push(delegateAddress);
            }
            else {
                throw new Error('Delegate address must be unique.');
            }
            const numberAmount = Number(amount);
            if (Number.isNaN(numberAmount)) {
                throw new error_1.ValidationError('Enter a valid amount in number string format.');
            }
            const sign = numberAmount < 0 ? -1 : 1;
            const normalizedAmount = lisk_transactions_1.utils.convertLSKToBeddows(String(Math.abs(numberAmount)));
            if (!lisk_validator_1.isValidFee(normalizedAmount)) {
                throw new error_1.ValidationError('Enter a valid vote amount in number string format.');
            }
            return {
                delegateAddress,
                amount: (BigInt(normalizedAmount) * BigInt(sign)).toString(),
            };
        });
        const validatedVotes = votesObjects ? validateAddresses(votesObjects) : [];
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        if (noSignature) {
            const noSignatureResult = processInputs(nonce, normalizedFee, networkIdentifier, validatedVotes);
            this.print(noSignatureResult);
            return;
        }
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await reader_1.getPassphraseFromPrompt('passphrase', true));
        const result = processInputs(nonce, normalizedFee, networkIdentifier, validatedVotes, passphrase);
        this.print(result);
    }
}
exports.default = VoteCommand;
VoteCommand.args = [
    {
        name: 'nonce',
        required: true,
        description: 'Nonce of the transaction.',
    },
    {
        name: 'fee',
        required: true,
        description: 'Transaction fee in LSK.',
    },
];
VoteCommand.description = `
	Creates a transaction which will cast votes for delegate candidates using their addresses if broadcast to the network.
	`;
VoteCommand.examples = [
    'transaction:create:vote 1 100 --votes="356975984361330918L,1000000000" --votes="7539210577161571444L,3000000000" --votes="456975984361330919L,-1000000000"',
];
VoteCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'no-signature': command_1.flags.boolean(flags_1.flags.noSignature),
    votes: command_1.flags.string({
        ...flags_1.flags.votes,
        multiple: true,
        required: true,
    }),
};
//# sourceMappingURL=vote.js.map