"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const lisk_transactions_1 = require("@liskhq/lisk-transactions");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../base"));
const error_1 = require("../../utils/error");
const flags_1 = require("../../utils/flags");
const network_identifier_1 = require("../../utils/network_identifier");
const object_1 = require("../../utils/object");
const reader_1 = require("../../utils/reader");
const transactions_1 = require("../../utils/transactions");
const getTransactionInput = async () => {
    try {
        const lines = await reader_1.readStdIn();
        if (!lines.length) {
            throw new error_1.ValidationError('No transaction was provided.');
        }
        return lines[0];
    }
    catch (e) {
        throw new error_1.ValidationError('No transaction was provided.');
    }
};
const getPassphrasesFromPrompt = async (numberOfPassphrases = 1) => {
    const passphrases = [];
    for (let index = 0; index < numberOfPassphrases; index += 1) {
        const passphrase = await reader_1.getPassphraseFromPrompt('passphrase', true);
        passphrases.push(passphrase);
    }
    return passphrases;
};
class SignCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'mandatory-key': mandatoryKeys, 'optional-key': optionalKeys, 'number-of-passphrases': numberOfPassphrases, }, } = this.parse(SignCommand);
        const { transaction } = args;
        const transactionInput = transaction || (await getTransactionInput());
        const transactionObject = transactions_1.parseTransactionString(transactionInput);
        const passphrase = passphraseSource !== null && passphraseSource !== void 0 ? passphraseSource : (await getPassphrasesFromPrompt(numberOfPassphrases));
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        const txInstance = transactions_1.instantiateTransaction({
            ...transactionObject,
            networkIdentifier,
        });
        const keys = {
            mandatoryKeys: mandatoryKeys !== null && mandatoryKeys !== void 0 ? mandatoryKeys : [],
            optionalKeys: optionalKeys !== null && optionalKeys !== void 0 ? optionalKeys : [],
        };
        if ((mandatoryKeys === null || mandatoryKeys === void 0 ? void 0 : mandatoryKeys.length) || (optionalKeys === null || optionalKeys === void 0 ? void 0 : optionalKeys.length)) {
            passphrase.forEach(p => {
                lisk_transactions_1.signMultiSignatureTransaction({
                    transaction: txInstance.toJSON(),
                    passphrase: p,
                    networkIdentifier,
                    keys,
                });
            });
        }
        else {
            txInstance.sign(networkIdentifier, passphrase[0]);
        }
        const { errors } = txInstance.validate();
        if (errors.length !== 0) {
            throw errors;
        }
        this.print(object_1.removeUndefinedValues(txInstance.toJSON()));
    }
}
exports.default = SignCommand;
SignCommand.args = [
    {
        name: 'transaction',
        description: 'Transaction to sign in JSON format.',
    },
];
SignCommand.description = `
	Sign a transaction using your secret passphrase.
	`;
SignCommand.examples = [
    'transaction:sign \'{"type":8,"senderPublicKey":"c094ebee7ec0","nonce":"1","fee":"1000","asset":{"amount":"100","recipientId":"555331L"}}\'',
    '\n',
    'transaction:sign \'{"type":8,"senderPublicKey":"c094ebee7ec0","nonce":"1","fee":"1000","asset":{"amount":"100","recipientId":"555331L"}}\' --mandatory-key=215b667a32a5cd51a94 --optional-key=922fbfdd596fa78269bbcadc67e --number-of-passphrases=2',
    '\n',
    'transaction:sign \'{"type":8,"senderPublicKey":"c094ebee7ec0","nonce":"1","fee":"1000","signatures":["a3cc97079e17bdd158526"],"asset":{"amount":"100","recipientId":"555331L"}}\' --mandatory-key=215b667a32a5cd51a94 --optional-key=922fbfdd596fa78269bbcadc67e --passphrase="inherit moon normal relief spring"',
    '\n',
    'transaction:sign \'{"type":8,"senderPublicKey":"c094ebee7ec0","nonce":"1","fee":"1000","asset":{"amount":"100","recipientId":"555331L"}}\' --mandatory-key=215b667a32a5cd51a94 --optional-key=922fbfdd596fa78269bbcadc67e --passphrase="inherit moon normal relief spring" --passphrase="wear protect skill sentence"',
];
SignCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    'mandatory-key': command_1.flags.string({
        ...flags_1.flags.mandatoryKey,
        multiple: true,
    }),
    'optional-key': command_1.flags.string({
        ...flags_1.flags.optionalKey,
        multiple: true,
    }),
    passphrase: command_1.flags.string({
        ...flags_1.flags.passphrase,
        multiple: true,
    }),
    'number-of-passphrases': command_1.flags.integer({
        ...flags_1.flags.numberOfPassphrases,
        exclusive: ['passphrase'],
    }),
};
//# sourceMappingURL=sign.js.map