"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const base_transaction_1 = require("./base_transaction");
const constants_1 = require("./constants");
const errors_1 = require("./errors");
const utils_1 = require("./utils");
const blockHeaderSchema = {
    type: 'object',
    required: [
        'blockSignature',
        'generatorPublicKey',
        'height',
        'maxHeightPreviouslyForged',
        'maxHeightPrevoted',
        'numberOfTransactions',
        'payloadHash',
        'payloadLength',
        'previousBlockId',
        'reward',
        'seedReveal',
        'timestamp',
        'totalAmount',
        'totalFee',
        'version',
    ],
    properties: {
        blockSignature: {
            type: 'string',
            format: 'signature',
        },
        generatorPublicKey: {
            type: 'string',
            format: 'publicKey',
        },
        height: {
            type: 'integer',
            minimum: 1,
        },
        maxHeightPreviouslyForged: {
            type: 'integer',
            minimum: 0,
        },
        maxHeightPrevoted: {
            type: 'integer',
            minimum: 0,
        },
        numberOfTransactions: {
            type: 'integer',
            minimum: 0,
        },
        payloadHash: {
            type: 'string',
            format: 'hex',
        },
        payloadLength: {
            type: 'integer',
            minimum: 0,
        },
        previousBlockId: {
            type: ['string'],
            format: 'id',
            minLength: 1,
            maxLength: 20,
        },
        reward: {
            type: 'string',
            format: 'amount',
        },
        seedReveal: {
            type: 'string',
            format: 'hex',
        },
        timestamp: {
            type: 'integer',
            minimum: 0,
        },
        totalAmount: {
            type: 'string',
            format: 'amount',
        },
        totalFee: {
            type: 'string',
            format: 'amount',
        },
        version: {
            type: 'integer',
            minimum: 0,
        },
    },
};
const proofOfMisbehaviorAssetFormatSchema = {
    type: 'object',
    required: ['header1', 'header2'],
    properties: {
        header1: blockHeaderSchema,
        header2: blockHeaderSchema,
    },
};
class ProofOfMisbehaviorTransaction extends base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        this.asset = (tx.asset || {});
        this.asset.reward =
            this.asset.reward && lisk_validator_1.isNumberString(this.asset.reward)
                ? BigInt(this.asset.reward)
                : BigInt(0);
    }
    assetToJSON() {
        return {
            header1: this.asset.header1,
            header2: this.asset.header2,
            reward: this.asset.reward.toString(),
        };
    }
    assetToBytes() {
        return Buffer.concat([
            utils_1.getBlockBytesWithSignature(this.asset.header1),
            utils_1.getBlockBytesWithSignature(this.asset.header2),
        ]);
    }
    async prepare(store) {
        const delegateAddress = lisk_cryptography_1.getAddressFromPublicKey(this.asset.header1.generatorPublicKey);
        const filterArray = [
            {
                address: this.senderId,
            },
            {
                address: delegateAddress,
            },
        ];
        await store.account.cache(filterArray);
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = lisk_validator_1.validator.validate(proofOfMisbehaviorAssetFormatSchema, asset);
        const errors = errors_1.convertToAssetError(this.id, schemaErrors);
        if (this.asset.header1.generatorPublicKey !==
            this.asset.header2.generatorPublicKey) {
            errors.push(new errors_1.TransactionError('GeneratorPublickey of each blockheader should match.', this.id, '.asset.header1.generatorPublicKey'));
        }
        if (Buffer.compare(utils_1.getBlockBytes(this.asset.header1), utils_1.getBlockBytes(this.asset.header2)) === 0) {
            errors.push(new errors_1.TransactionError('Blockheaders are identical. No contradiction detected.', this.id, '.asset.header1'));
        }
        let b1 = this.asset.header1;
        let b2 = this.asset.header2;
        if (b1.maxHeightPreviouslyForged > b2.maxHeightPreviouslyForged ||
            (b1.maxHeightPreviouslyForged === b2.maxHeightPreviouslyForged &&
                b1.maxHeightPrevoted > b2.maxHeightPrevoted) ||
            (b1.maxHeightPreviouslyForged === b2.maxHeightPreviouslyForged &&
                b1.maxHeightPrevoted === b2.maxHeightPrevoted &&
                b1.height > b2.height)) {
            b1 = this.asset.header2;
            b2 = this.asset.header1;
        }
        if (!(b1.maxHeightPrevoted === b2.maxHeightPrevoted && b1.height >= b2.height) &&
            !(b1.height > b2.maxHeightPreviouslyForged) &&
            !(b1.maxHeightPrevoted > b2.maxHeightPrevoted)) {
            errors.push(new errors_1.TransactionError('Blockheaders are not contradicting as per BFT violation rules.', this.id, '.asset.header1'));
        }
        return errors;
    }
    async applyAsset(store) {
        const errors = [];
        const currentHeight = store.chain.lastBlockHeader.height + 1;
        const senderAccount = await store.account.get(this.senderId);
        const { networkIdentifier } = store.chain;
        if (Math.abs(this.asset.header1.height - currentHeight) >=
            constants_1.MAX_PUNISHABLE_BLOCK_HEIGHT_DIFFERENCE) {
            errors.push(new errors_1.TransactionError(`Difference between header1.height and current height must be less than ${constants_1.MAX_PUNISHABLE_BLOCK_HEIGHT_DIFFERENCE}.`, this.id, '.asset.header1', this.asset.header1.height));
        }
        if (Math.abs(this.asset.header2.height - currentHeight) >=
            constants_1.MAX_PUNISHABLE_BLOCK_HEIGHT_DIFFERENCE) {
            errors.push(new errors_1.TransactionError(`Difference between header2.height and current height must be less than ${constants_1.MAX_PUNISHABLE_BLOCK_HEIGHT_DIFFERENCE}.`, this.id, '.asset.header2', this.asset.header2.height));
        }
        const delegateAddress = lisk_cryptography_1.getAddressFromPublicKey(this.asset.header1.generatorPublicKey);
        const delegateAccount = await store.account.getOrDefault(delegateAddress);
        if (!delegateAccount.isDelegate || !delegateAccount.username) {
            errors.push(new errors_1.TransactionError('Account is not a delegate', this.id, '.asset.header1.generatorPublicKey'));
            return errors;
        }
        if (delegateAccount.delegate.isBanned) {
            errors.push(new errors_1.TransactionError('Cannot apply proof-of-misbehavior. Delegate is banned.', this.id, '.asset.header1.generatorPublicKey', this.asset.header1.generatorPublicKey));
            return errors;
        }
        if (utils_1.getPunishmentPeriod(delegateAccount, delegateAccount, store.chain.lastBlockHeader.height) > 0) {
            errors.push(new errors_1.TransactionError('Cannot apply proof-of-misbehavior. Delegate is already punished. ', this.id, '.asset.header1.generatorPublicKey', this.asset.header1.generatorPublicKey));
            return errors;
        }
        const blockHeader1Bytes = Buffer.concat([
            Buffer.from(networkIdentifier, 'hex'),
            utils_1.getBlockBytes(this.asset.header1),
        ]);
        const blockHeader2Bytes = Buffer.concat([
            Buffer.from(networkIdentifier, 'hex'),
            utils_1.getBlockBytes(this.asset.header2),
        ]);
        const { valid: validHeader1Signature } = utils_1.validateSignature(this.asset.header1.generatorPublicKey, this.asset.header1.blockSignature, blockHeader1Bytes);
        if (!validHeader1Signature) {
            errors.push(new errors_1.TransactionError('Invalid block signature for header 1.', this.id, '.asset.header1.blockSignature', this.asset.header1.blockSignature));
        }
        const { valid: validHeader2Signature } = utils_1.validateSignature(this.asset.header2.generatorPublicKey, this.asset.header2.blockSignature, blockHeader2Bytes);
        if (!validHeader2Signature) {
            errors.push(new errors_1.TransactionError('Invalid block signature for header 2.', this.id, '.asset.header2.blockSignature', this.asset.header2.blockSignature));
        }
        const reward = store.chain.lastBlockReward > delegateAccount.balance
            ? delegateAccount.balance
            : store.chain.lastBlockReward;
        senderAccount.balance += reward;
        this.asset.reward = reward;
        store.account.set(senderAccount.address, senderAccount);
        const updatedDelegateAccount = await store.account.get(delegateAddress);
        updatedDelegateAccount.delegate.pomHeights.push(currentHeight);
        if (updatedDelegateAccount.delegate.pomHeights.length >= constants_1.MAX_POM_HEIGHTS) {
            updatedDelegateAccount.delegate.isBanned = true;
        }
        updatedDelegateAccount.balance -= reward;
        store.account.set(updatedDelegateAccount.address, updatedDelegateAccount);
        return errors;
    }
    async undoAsset(store) {
        const currentHeight = store.chain.lastBlockHeader.height + 1;
        const senderAccount = await store.account.get(this.senderId);
        senderAccount.balance -= this.asset.reward;
        store.account.set(senderAccount.address, senderAccount);
        const delegateAddress = lisk_cryptography_1.getAddressFromPublicKey(this.asset.header1.generatorPublicKey);
        const delegateAccount = await store.account.get(delegateAddress);
        const pomIndex = delegateAccount.delegate.pomHeights.findIndex(height => height === currentHeight);
        delegateAccount.delegate.pomHeights.splice(pomIndex, 1);
        if (delegateAccount.delegate.pomHeights.length < 5) {
            delegateAccount.delegate.isBanned = false;
        }
        delegateAccount.balance += this.asset.reward;
        store.account.set(delegateAccount.address, delegateAccount);
        return [];
    }
}
exports.ProofOfMisbehaviorTransaction = ProofOfMisbehaviorTransaction;
ProofOfMisbehaviorTransaction.TYPE = 15;
//# sourceMappingURL=15_proof_of_misbehavior_transaction.js.map