"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const constants_1 = require("./constants");
const errors_1 = require("./errors");
const response_1 = require("./response");
const schemas = require("./schema");
const utils_1 = require("./utils");
exports.ENTITY_ACCOUNT = 'account';
exports.ENTITY_TRANSACTION = 'transaction';
class BaseTransaction {
    constructor(rawTransaction) {
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        this.senderPublicKey = tx.senderPublicKey || '';
        this.signatures = tx.signatures || [];
        this.nonce =
            tx.nonce && lisk_validator_1.isValidNonce(tx.nonce) ? BigInt(tx.nonce) : BigInt(0);
        this.fee = tx.fee && lisk_validator_1.isValidFee(tx.fee) ? BigInt(tx.fee) : BigInt(0);
        this.type =
            typeof tx.type === 'number'
                ? tx.type
                : this.constructor.TYPE;
        this._id = tx.id;
        this.confirmations = tx.confirmations;
        this.blockId = tx.blockId;
        this.height = tx.height;
        this.receivedAt = tx.receivedAt ? new Date(tx.receivedAt) : undefined;
        this.asset = tx.asset || {};
    }
    get id() {
        return this._id || 'incalculable-id';
    }
    get minFee() {
        if (!this._minFee) {
            this._minFee =
                this.constructor.NAME_FEE +
                    BigInt(this.constructor.MIN_FEE_PER_BYTE) *
                        BigInt(this.getBytes().length);
        }
        return this._minFee;
    }
    get senderId() {
        return lisk_cryptography_1.getAddressFromPublicKey(this.senderPublicKey);
    }
    toJSON() {
        const transaction = {
            id: this._id,
            blockId: this.blockId,
            height: this.height,
            confirmations: this.confirmations,
            type: this.type,
            senderPublicKey: this.senderPublicKey,
            senderId: this.senderPublicKey ? this.senderId : '',
            nonce: this.nonce.toString(),
            fee: this.fee.toString(),
            signatures: this.signatures,
            asset: this.assetToJSON(),
            receivedAt: this.receivedAt ? this.receivedAt.toISOString() : undefined,
        };
        return transaction;
    }
    stringify() {
        return JSON.stringify(this.toJSON());
    }
    getBytes() {
        const transactionBytes = Buffer.concat([
            this.getBasicBytes(),
            utils_1.serializeSignatures(this.signatures),
        ]);
        return transactionBytes;
    }
    validate() {
        const errors = [...this._validateSchema(), ...this.validateAsset()];
        if (errors.length > 0) {
            return response_1.createResponse(this.id, errors);
        }
        this._id = utils_1.getId(this.getBytes());
        if (this.type !== this.constructor.TYPE) {
            errors.push(new errors_1.TransactionError(`Invalid transaction type`, this.id, '.type', this.type, this.constructor.TYPE));
        }
        if (this.fee < this.minFee) {
            errors.push(new errors_1.TransactionError(`Insufficient transaction fee. Minimum required fee is: ${this.minFee.toString()}`, this.id, '.fee', this.fee.toString()));
        }
        return response_1.createResponse(this.id, errors);
    }
    async apply(store) {
        const sender = await store.account.getOrDefault(this.senderId);
        const errors = [];
        const senderPublicKeyError = utils_1.verifySenderPublicKey(this.id, sender, this.senderPublicKey);
        if (senderPublicKeyError) {
            errors.push(senderPublicKeyError);
        }
        const accountNonceError = utils_1.verifyAccountNonce(this.id, sender, this.nonce);
        if (accountNonceError) {
            errors.push(accountNonceError);
        }
        const { errors: signaturesErr } = await this.verifySignatures(store);
        if (signaturesErr) {
            errors.push(...signaturesErr);
        }
        sender.balance -= this.fee;
        sender.publicKey = sender.publicKey || this.senderPublicKey;
        sender.nonce += BigInt(1);
        store.account.set(sender.address, sender);
        const assetErrors = await this.applyAsset(store);
        errors.push(...assetErrors);
        const updatedSender = await store.account.get(this.senderId);
        const minRemainingBalanceError = utils_1.verifyMinRemainingBalance(this.id, updatedSender, this.constructor.MIN_REMAINING_BALANCE);
        if (minRemainingBalanceError) {
            errors.push(minRemainingBalanceError);
        }
        return response_1.createResponse(this.id, errors);
    }
    async undo(store) {
        const sender = await store.account.getOrDefault(this.senderId);
        const updatedBalance = sender.balance + this.fee;
        sender.balance = updatedBalance;
        sender.publicKey = sender.publicKey || this.senderPublicKey;
        const errors = updatedBalance <= BigInt(constants_1.MAX_TRANSACTION_AMOUNT)
            ? []
            : [
                new errors_1.TransactionError('Invalid balance amount', this.id, '.balance', sender.balance.toString(), updatedBalance.toString()),
            ];
        sender.nonce -= BigInt(1);
        store.account.set(sender.address, sender);
        const assetErrors = await this.undoAsset(store);
        errors.push(...assetErrors);
        return response_1.createResponse(this.id, errors);
    }
    async prepare(store) {
        await store.account.cache([
            {
                address: this.senderId,
            },
        ]);
    }
    async verifySignatures(store) {
        const sender = await store.account.get(this.senderId);
        const { networkIdentifier } = store.chain;
        const transactionBytes = this.getBasicBytes();
        if (networkIdentifier === undefined || networkIdentifier === '') {
            throw new Error('Network identifier is required to validate a transaction ');
        }
        const networkIdentifierBytes = lisk_cryptography_1.hexToBuffer(networkIdentifier);
        const transactionWithNetworkIdentifierBytes = Buffer.concat([
            networkIdentifierBytes,
            transactionBytes,
        ]);
        if (!utils_1.isMultisignatureAccount(sender)) {
            const { error } = utils_1.validateSignature(this.senderPublicKey, this.signatures[0], transactionWithNetworkIdentifierBytes, this.id);
            if (error) {
                return response_1.createResponse(this.id, [error]);
            }
            return response_1.createResponse(this.id, []);
        }
        const errors = utils_1.verifyMultiSignatureTransaction(this.id, sender, this.signatures, transactionWithNetworkIdentifierBytes);
        return response_1.createResponse(this.id, errors);
    }
    sign(networkIdentifier, senderPassphrase, passphrases, keys) {
        if (!networkIdentifier) {
            throw new Error('Network identifier is required to sign a transaction');
        }
        const networkIdentifierBytes = lisk_cryptography_1.hexToBuffer(networkIdentifier);
        if (senderPassphrase) {
            const { publicKey } = lisk_cryptography_1.getAddressAndPublicKeyFromPassphrase(senderPassphrase);
            if (this.senderPublicKey !== '' && this.senderPublicKey !== publicKey) {
                throw new Error('Transaction senderPublicKey does not match public key from passphrase');
            }
            this.senderPublicKey = publicKey;
            const transactionWithNetworkIdentifierBytes = Buffer.concat([
                networkIdentifierBytes,
                this.getBasicBytes(),
            ]);
            const signature = lisk_cryptography_1.signData(lisk_cryptography_1.hash(transactionWithNetworkIdentifierBytes), senderPassphrase);
            this.signatures = [];
            this.signatures.push(signature);
            this._id = utils_1.getId(this.getBytes());
            return;
        }
        if (passphrases && keys) {
            if (!this.senderPublicKey) {
                throw new Error('Transaction senderPublicKey needs to be set before signing');
            }
            const transactionWithNetworkIdentifierBytes = Buffer.concat([
                networkIdentifierBytes,
                this.getBasicBytes(),
            ]);
            const keysAndPassphrases = utils_1.buildPublicKeyPassphraseDict(passphrases);
            utils_1.sortKeysAscending(keys.mandatoryKeys);
            utils_1.sortKeysAscending(keys.optionalKeys);
            for (const aKey of [...keys.mandatoryKeys, ...keys.optionalKeys]) {
                if (keysAndPassphrases[aKey]) {
                    const { passphrase } = keysAndPassphrases[aKey];
                    this.signatures.push(lisk_cryptography_1.signData(lisk_cryptography_1.hash(transactionWithNetworkIdentifierBytes), passphrase));
                }
                else {
                    this.signatures.push('');
                }
            }
            this._id = utils_1.getId(this.getBytes());
            return;
        }
        return;
    }
    getBasicBytes() {
        const transactionType = Buffer.alloc(constants_1.BYTESIZES.TYPE, this.type);
        const transactionNonce = lisk_cryptography_1.intToBuffer(this.nonce.toString(), constants_1.BYTESIZES.NONCE);
        const transactionSenderPublicKey = lisk_cryptography_1.hexToBuffer(this.senderPublicKey);
        const transactionFee = lisk_cryptography_1.intToBuffer(this.fee.toString(), constants_1.BYTESIZES.FEE);
        return Buffer.concat([
            transactionType,
            transactionNonce,
            transactionSenderPublicKey,
            transactionFee,
            this.assetToBytes(),
        ]);
    }
    assetToJSON() {
        return this.asset;
    }
    assetToBytes() {
        return Buffer.from(JSON.stringify(this.asset), 'utf-8');
    }
    _validateSchema() {
        const transaction = this.toJSON();
        const schemaErrors = lisk_validator_1.validator.validate(schemas.baseTransaction, transaction);
        const errors = errors_1.convertToTransactionError(this.id, schemaErrors);
        if (!errors.find((err) => err.dataPath === '.senderPublicKey')) {
            const senderIdError = utils_1.validateSenderIdAndPublicKey(this.id, this.senderId, this.senderPublicKey);
            if (senderIdError) {
                errors.push(senderIdError);
            }
        }
        return errors;
    }
}
exports.BaseTransaction = BaseTransaction;
BaseTransaction.MIN_REMAINING_BALANCE = BigInt('5000000');
BaseTransaction.MIN_FEE_PER_BYTE = constants_1.MIN_FEE_PER_BYTE;
BaseTransaction.NAME_FEE = BigInt(0);
//# sourceMappingURL=base_transaction.js.map